# -*- coding: utf-8 -*-

import logging
from threading import Event, Thread


class StoppableThread(Thread):
    """Thread class with a stop() method. The thread itself has to check
    regularly for the stopped() condition."""

    def __init__(self, **kwargs):
        self.logger = kwargs.pop("logger", logging.getLogger(__name__))
        self._stop_event = Event()

        Thread.__init__(self, **kwargs)

    def stop(self):
        """Sets the stop event"""
        self.logger.debug("Stopping thread: %s", self.name)
        self._stop_event.set()

    def stopped(self):
        """Determines if stop has been called yet."""
        return self._stop_event.isSet()

    def wait(self, timeout=None):
        """Delegate wait call to threading.Event"""
        return self._stop_event.wait(timeout)



# -*- coding: utf-8 -*-

import json
import logging

from six import string_types


class BrewtilsException(Exception):
    """Base exception"""

    pass


# Error Logging Control
class SuppressStacktrace(Exception):
    """Mixin that will suppress stacktrace logging"""

    _bg_suppress_stacktrace = True


class ErrorLogLevelCritical(Exception):
    """Mixin to log an exception at the CRITICAL level"""

    _bg_error_log_level = logging.CRITICAL


class ErrorLogLevelError(Exception):
    """Mixin to log an exception at the ERROR level"""

    _bg_error_log_level = logging.ERROR


class ErrorLogLevelWarning(Exception):
    """Mixin to log an exception at the WARNING level"""

    _bg_error_log_level = logging.WARNING


class ErrorLogLevelInfo(Exception):
    """Mixin to log an exception at the INFO level"""

    _bg_error_log_level = logging.INFO


class ErrorLogLevelDebug(Exception):
    """Mixin to log an exception at the DEBUG level"""

    _bg_error_log_level = logging.DEBUG


# Models
class ModelError(BrewtilsException):
    """Base exception for model errors"""

    pass


class ModelValidationError(ModelError):
    """Invalid model"""

    pass


class RequestStatusTransitionError(ModelValidationError):
    """A status update was an invalid transition"""

    pass


# Plugins
class PluginError(BrewtilsException):
    """Generic error class"""

    pass


class PluginValidationError(PluginError):
    """Plugin could not be validated successfully"""

    pass


class PluginParamError(PluginError):
    """Error used when plugins have illegal parameters"""

    pass


# Requests
class RequestProcessException(BrewtilsException):
    """Base for exceptions that occur during request processing"""

    pass


class AckAndContinueException(RequestProcessException):
    pass


class NoAckAndDieException(RequestProcessException):
    pass


class AckAndDieException(RequestProcessException):
    pass


class DiscardMessageException(RequestProcessException):
    """Raising an instance will result in a message not being requeued"""

    pass


class RepublishRequestException(RequestProcessException):
    """Republish to the end of the message queue

    :param request: The Request to republish
    :param headers: A dictionary of headers to be used by
        `brewtils.request_consumer.RequestConsumer`
    :type request: :py:class:`brewtils.models.Request`
    """

    def __init__(self, request, headers):
        self.request = request
        self.headers = headers


class RequestProcessingError(AckAndContinueException):
    pass


class RequestPublishException(BrewtilsException):
    """Error while publishing request"""

    pass


# Rest / Client errors
class RestError(BrewtilsException):
    """Base exception for REST errors"""

    pass


class RestClientError(RestError):
    """Wrapper for all 4XX errors"""

    pass


class RestServerError(RestError):
    """Wrapper for all 5XX errors"""

    pass


class RestConnectionError(RestServerError):
    """Error indicating a connection error while performing a request"""

    pass


class FetchError(RestError):
    """Error Indicating a server Error occurred performing a GET"""

    pass


class ValidationError(RestClientError):
    """Error Indicating a client (400) Error occurred performing a POST/PUT"""

    pass


class SaveError(RestServerError):
    """Error Indicating a server Error occurred performing a POST/PUT"""

    pass


class DeleteError(RestServerError):
    """Error Indicating a server Error occurred performing a DELETE"""

    pass


class TimeoutExceededError(RestClientError):
    """Error indicating a timeout occurred waiting for a request to complete"""

    pass


class ConflictError(RestClientError):
    """Error indicating a 409 was raised on the server"""

    pass


class RequestFailedError(RestError):
    """Request returned with a 200, but the status was ERROR"""

    def __init__(self, request):
        self.request = request


class NotFoundError(RestClientError):
    """Error Indicating a 404 was raised on the server"""

    pass


class RequestForbidden(RestClientError):
    """Error indicating a 403 was raised on the server"""

    pass


class AuthorizationRequired(RestClientError):
    """Error indicating a 401 was raised on the server"""

    pass


# Alias old names
WaitExceededError = TimeoutExceededError
ConnectionTimeoutError = TimeoutExceededError

BrewmasterModelError = ModelError
BrewmasterModelValidationError = ModelValidationError
BrewmasterRestError = RestError
BrewmasterRestClientError = RestClientError
BrewmasterRestServerError = RestServerError
BrewmasterConnectionError = RestConnectionError
BrewmasterTimeoutError = ConnectionTimeoutError
BrewmasterFetchError = FetchError
BrewmasterValidationError = ValidationError
BrewmasterSaveError = SaveError
BrewmasterDeleteError = DeleteError
BGConflictError = ConflictError
BGRequestFailedError = RequestFailedError
BGNotFoundError = NotFoundError


def parse_exception_as_json(exc):
    """
    Attempt to parse an Exception to a JSON string.

    If the exception has a single argument, no attributes, and the attribute
    can be converted to a valid JSON string, then that will be returned.

    Otherwise, a string version of the following form will be returned:

    {
        "message": "",
        "arguments": [],
        "attributes": {}
    }

    Where "message" is just str(exc), "arguments" is a list of all the
    arguments passed to the exception attempted to be converted to a valid
    JSON string, and "attributes" are the attributes of the exception class.

    If parsing fails at all, then a simple str() will be applied either the
    argument or attribute value.

    Note:
        On python version 2, errors with custom attributes do not list those
        attributes as arguments.

    Args:
        exc (Exception): The exception you would like to format as JSON.

    Raises:
        ValueError: If the exception passed in is not an Exception.

    Returns:
        A valid JSON string representing (the best we can) the exception.

    """
    if not isinstance(exc, Exception):
        raise ValueError("Attempted to parse a non-exception as JSON.")

    json_args = []
    valid_json = True
    for arg in exc.args:
        valid_json, json_arg = _jsonify_value(arg)
        json_args.append(json_arg)

    if (
        len(json_args) == 1
        and not exc.__dict__
        and valid_json
        and isinstance(json_args[0], (list, dict))
    ):
        return json.dumps(json_args[0])

    return json.dumps(
        {
            "message": str(exc),
            "arguments": json_args,
            "attributes": _jsonify_value(exc.__dict__)[1],
        }
    )


def _jsonify_value(value):
    """Attempt to JSONify a value, returns success and then a string"""
    try:
        if isinstance(value, string_types):
            v = json.loads(value)
            if isinstance(v, string_types):
                return True, value
            else:
                return True, v
        else:
            json.dumps(value)
            return True, value
    except Exception:
        return False, str(value)



# -*- coding: utf-8 -*-

import logging
import warnings

import six

from brewtils.models import (
    System,
    Instance,
    Command,
    Parameter,
    Request,
    PatchOperation,
    Choices,
    LoggingConfig,
    Event,
    Queue,
    Principal,
    Role,
    RefreshToken,
    Job,
    RequestTemplate,
    DateTrigger,
    CronTrigger,
    IntervalTrigger,
)
from brewtils.schemas import (
    SystemSchema,
    InstanceSchema,
    CommandSchema,
    ParameterSchema,
    RequestSchema,
    PatchSchema,
    LoggingConfigSchema,
    EventSchema,
    QueueSchema,
    PrincipalSchema,
    RoleSchema,
    RefreshTokenSchema,
    JobSchema,
)


class SchemaParser(object):
    """Serialize and deserialize Brewtils models"""

    _models = {
        "SystemSchema": System,
        "InstanceSchema": Instance,
        "CommandSchema": Command,
        "ParameterSchema": Parameter,
        "RequestTemplateSchema": RequestTemplate,
        "RequestSchema": Request,
        "PatchSchema": PatchOperation,
        "ChoicesSchema": Choices,
        "LoggingConfigSchema": LoggingConfig,
        "EventSchema": Event,
        "QueueSchema": Queue,
        "PrincipalSchema": Principal,
        "RoleSchema": Role,
        "RefreshTokenSchema": RefreshToken,
        "JobSchema": Job,
        "DateTriggerSchema": DateTrigger,
        "IntervalTriggerSchema": IntervalTrigger,
        "CronTriggerSchema": CronTrigger,
    }

    logger = logging.getLogger(__name__)

    # Deserialization methods
    @classmethod
    def parse_system(cls, system, from_string=False, **kwargs):
        """Convert raw JSON string or dictionary to a system model object

        :param system: The raw input
        :param from_string: True if input is a JSON string, False if a dictionary
        :param kwargs: Additional parameters to be passed to the Schema (e.g. many=True)
        :return: A System object
        """
        return cls._do_parse(system, SystemSchema(**kwargs), from_string=from_string)

    @classmethod
    def parse_instance(cls, instance, from_string=False, **kwargs):
        """Convert raw JSON string or dictionary to an instance model object

        :param instance: The raw input
        :param from_string: True if input is a JSON string, False if a dictionary
        :param kwargs: Additional parameters to be passed to the Schema (e.g. many=True)
        :return: An Instance object
        """
        return cls._do_parse(
            instance, InstanceSchema(**kwargs), from_string=from_string
        )

    @classmethod
    def parse_command(cls, command, from_string=False, **kwargs):
        """Convert raw JSON string or dictionary to a command model object

        :param command: The raw input
        :param from_string: True if input is a JSON string, False if a dictionary
        :param kwargs: Additional parameters to be passed to the Schema (e.g. many=True)
        :return: A Command object
        """
        return cls._do_parse(command, CommandSchema(**kwargs), from_string=from_string)

    @classmethod
    def parse_parameter(cls, parameter, from_string=False, **kwargs):
        """Convert raw JSON string or dictionary to a parameter model object

        :param parameter: The raw input
        :param from_string: True if input is a JSON string, False if a dictionary
        :param kwargs: Additional parameters to be passed to the Schema (e.g. many=True)
        :return: An Parameter object
        """
        return cls._do_parse(
            parameter, ParameterSchema(**kwargs), from_string=from_string
        )

    @classmethod
    def parse_request(cls, request, from_string=False, **kwargs):
        """Convert raw JSON string or dictionary to a request model object

        :param request: The raw input
        :param from_string: True if input is a JSON string, False if a dictionary
        :param kwargs: Additional parameters to be passed to the Schema (e.g. many=True)
        :return: A Request object
        """
        return cls._do_parse(request, RequestSchema(**kwargs), from_string=from_string)

    @classmethod
    def parse_patch(cls, patch, from_string=False, **kwargs):
        """Convert raw JSON string or dictionary to a patch model object

        Note: for our patches, many is _always_ set to True. We will always return a list
        from this method.

        :param patch: The raw input
        :param from_string: True if input is a JSON string, False if a dictionary
        :param kwargs: Additional parameters to be passed to the Schema (e.g. many=True)
        :return: A PatchOperation object
        """
        if not kwargs.pop("many", True):
            cls.logger.warning(
                "A patch object should always be wrapped as a list of objects. "
                "Thus, parsing will always return a list. You specified many as "
                "False, this is being ignored and a list "
                "will be returned anyway."
            )
        return cls._do_parse(
            patch, PatchSchema(many=True, **kwargs), from_string=from_string
        )

    @classmethod
    def parse_logging_config(cls, logging_config, from_string=False, **kwargs):
        """Convert raw JSON string or dictionary to a logging config model object

        Note: for our logging_config, many is _always_ set to False. We will always return a dict
        from this method.

        :param logging_config: The raw input
        :param from_string: True if 'input is a JSON string, False if a dictionary
        :param kwargs: Additional parameters to be passed to the Schema (e.g. many=True)
        :return: A LoggingConfig object
        """
        if kwargs.pop("many", False):
            cls.logger.warning(
                "A logging config object should never be wrapped as a list of "
                "objects. Thus, parsing will always return a dict. You specified "
                "many as True, this is being ignored and a dict will be returned "
                "anyway."
            )
        return cls._do_parse(
            logging_config,
            LoggingConfigSchema(many=False, **kwargs),
            from_string=from_string,
        )

    @classmethod
    def parse_event(cls, event, from_string=False, **kwargs):
        """Convert raw JSON string or dictionary to an event model object

        :param event: The raw input
        :param from_string: True if input is a JSON string, False if a dictionary
        :param kwargs: Additional parameters to be passed to the Schema (e.g. many=True)
        :return: An Event object
        """
        return cls._do_parse(event, EventSchema(**kwargs), from_string=from_string)

    @classmethod
    def parse_queue(cls, queue, from_string=False, **kwargs):
        """Convert raw JSON string or dictionary to a queue model object

        :param queue: The raw input
        :param from_string: True if input is a JSON string, False if a dictionary
        :param kwargs: Additional parameters to be passed to the Schema (e.g. many=True)
        :return: A Queue object
        """
        return cls._do_parse(queue, QueueSchema(**kwargs), from_string=from_string)

    @classmethod
    def parse_principal(cls, principal, from_string=False, **kwargs):
        """Convert raw JSON string or dictionary to a principal model object

        :param principal: The raw input
        :param from_string: True if input is a JSON string, False if a dictionary
        :param kwargs: Additional parameters to be passed to the Schema (e.g. many=True)
        :return: A Principal object
        """
        return cls._do_parse(
            principal, PrincipalSchema(**kwargs), from_string=from_string
        )

    @classmethod
    def parse_role(cls, role, from_string=False, **kwargs):
        """Convert raw JSON string or dictionary to a role model object

        :param role: The raw input
        :param from_string: True if input is a JSON string, False if a dictionary
        :param kwargs: Additional parameters to be passed to the Schema (e.g. many=True)
        :return: A Role object
        """
        return cls._do_parse(role, RoleSchema(**kwargs), from_string=from_string)

    @classmethod
    def parse_refresh_token(cls, refresh_token, from_string=False, **kwargs):
        """Convert raw JSON string or dictionary to a refresh token object

        :param refresh_token: The raw input
        :param from_string: True if input is a JSON string, False if a dictionary
        :param kwargs: Additional parameters to be passed to the Schema (e.g. many=True)
        :return: A RefreshToken object
        """
        return cls._do_parse(
            refresh_token, RefreshTokenSchema(**kwargs), from_string=from_string
        )

    @classmethod
    def parse_job(cls, job, from_string=False, **kwargs):
        """Convert raw JSON string or dictionary to a job model object

        Args:
            job: Raw input
            from_string: True if input is a JSON string, False if a dictionary
            **kwargs: Additional parameters to be passed to the Schema (e.g. many=True)

        Returns:
            A Job object.

        """
        return cls._do_parse(job, JobSchema(**kwargs), from_string=from_string)

    @classmethod
    def _do_parse(cls, data, schema, from_string=False):
        if from_string and not isinstance(data, six.string_types):
            raise TypeError("When from_string=True data must be a string-type")

        schema.context["models"] = cls._models
        return schema.loads(data).data if from_string else schema.load(data).data

    # Serialization methods
    @classmethod
    def serialize_system(cls, system, to_string=True, include_commands=True, **kwargs):
        """Convert a system model into serialized form

        :param system: The system object(s) to be serialized
        :param to_string: True to generate a JSON-formatted string, False to generate a dictionary
        :param include_commands: True if the system's command list should be included
        :param kwargs: Additional parameters to be passed to the Schema (e.g. many=True)
        :return: Serialized representation of system
        """
        if not include_commands:
            if "exclude" in kwargs:
                kwargs["exclude"] += ("commands",)
            else:
                kwargs["exclude"] = ("commands",)

        return cls._do_serialize(SystemSchema(**kwargs), system, to_string)

    @classmethod
    def serialize_instance(cls, instance, to_string=True, **kwargs):
        """Convert an instance model into serialized form

        :param instance: The instance object(s) to be serialized
        :param to_string: True to generate a JSON-formatted string, False to generate a dictionary
        :param kwargs: Additional parameters to be passed to the Schema (e.g. many=True)
        :return: Serialized representation of instance
        """
        return cls._do_serialize(InstanceSchema(**kwargs), instance, to_string)

    @classmethod
    def serialize_command(cls, command, to_string=True, **kwargs):
        """Convert a command model into serialized form

        :param command: The command object(s) to be serialized
        :param to_string: True to generate a JSON-formatted string, False to generate a dictionary
        :param kwargs: Additional parameters to be passed to the Schema (e.g. many=True)
        :return: Serialized representation of command
        """
        return cls._do_serialize(CommandSchema(**kwargs), command, to_string)

    @classmethod
    def serialize_parameter(cls, parameter, to_string=True, **kwargs):
        """Convert a parameter model into serialized form

        :param parameter: The parameter object(s) to be serialized
        :param to_string: True to generate a JSON-formatted string, False to generate a dictionary
        :param kwargs: Additional parameters to be passed to the Schema (e.g. many=True)
        :return: Serialized representation of parameter
        """
        return cls._do_serialize(ParameterSchema(**kwargs), parameter, to_string)

    @classmethod
    def serialize_request(cls, request, to_string=True, **kwargs):
        """Convert a request model into serialized form

        :param request: The request object(s) to be serialized
        :param to_string: True to generate a JSON-formatted string, False to generate a dictionary
        :param kwargs: Additional parameters to be passed to the Schema (e.g. many=True)
        :return: Serialized representation of request
        """
        return cls._do_serialize(RequestSchema(**kwargs), request, to_string)

    @classmethod
    def serialize_patch(cls, patch, to_string=True, **kwargs):
        """Convert a patch model into serialized form

        :param patch: The patch object(s) to be serialized
        :param to_string: True to generate a JSON-formatted string, False to generate a dictionary
        :param kwargs: Additional parameters to be passed to the Schema (e.g. many=True)
        :return: Serialized representation of patch
        """
        return cls._do_serialize(PatchSchema(**kwargs), patch, to_string)

    @classmethod
    def serialize_logging_config(cls, logging_config, to_string=True, **kwargs):
        """Convert a logging config model into serialize form

        :param logging_config: The logging config object(s) to be serialized
        :param to_string: True to generate a JSON-formatted string, False to generate a dictionary
        :param kwargs: Additional parameters to be passed to the Schema (e.g. many=True)
        :return: Serialized representation of logging config
        """
        return cls._do_serialize(
            LoggingConfigSchema(**kwargs), logging_config, to_string
        )

    @classmethod
    def serialize_event(cls, event, to_string=True, **kwargs):
        """Convert a logging config model into serialized form

        :param event: The event object(s) to be serialized
        :param to_string: True to generate a JSON-formatted string, False to generate a dictionary
        :param kwargs: Additional parameters to be passed to the Schema (e.g. many=True)
        :return: Serialized representation of event
        """
        return cls._do_serialize(EventSchema(**kwargs), event, to_string)

    @classmethod
    def serialize_queue(cls, queue, to_string=True, **kwargs):
        """Convert a queue model into serialized form

        :param queue: The queue object(s) to be serialized
        :param to_string: True to generate a JSON-formatted string, False to generate a dictionary
        :param kwargs: Additional parameters to be passed to the Schema (e.g. many=True)
        :return: Serialized representation of queue
        """
        return cls._do_serialize(QueueSchema(**kwargs), queue, to_string)

    @classmethod
    def serialize_principal(cls, principal, to_string=True, **kwargs):
        """Convert a principal model into serialized form

        :param principal: The principal object(s) to be serialized
        :param to_string: True to generate a JSON-formatted string, False to generate a dictionary
        :param kwargs: Additional parameters to be passed to the Schema (e.g. many=True)
        :return: Serialized representation
        """
        return cls._do_serialize(PrincipalSchema(**kwargs), principal, to_string)

    @classmethod
    def serialize_role(cls, role, to_string=True, **kwargs):
        """Convert a role model into serialized form

        :param role: The role object(s) to be serialized
        :param to_string: True to generate a JSON-formatted string, False to generate a dictionary
        :param kwargs: Additional parameters to be passed to the Schema (e.g. many=True)
        :return: Serialized representation
        """
        return cls._do_serialize(RoleSchema(**kwargs), role, to_string)

    @classmethod
    def serialize_refresh_token(cls, refresh_token, to_string=True, **kwargs):
        """Convert a role model into serialized form

        :param refresh_token: The token object(s) to be serialized
        :param to_string: True to generate a JSON-formatted string, False to generate a dictionary
        :param kwargs: Additional parameters to be passed to the Schema (e.g. many=True)
        :return: Serialized representation
        """
        return cls._do_serialize(RefreshTokenSchema(**kwargs), refresh_token, to_string)

    @classmethod
    def serialize_job(cls, job, to_string=True, **kwargs):
        """Convert a job model into serialized form.

        Args:
            job: The job object(s) to be serialized.
            to_string: True to generate a JSON-formatted string, False to generate a dictionary.
            **kwargs: Additional parameters to be passed to the shcema (e.g. many=True)

        Returns:
            Serialize representation of job.
        """
        return cls._do_serialize(JobSchema(**kwargs), job, to_string)

    @staticmethod
    def _do_serialize(schema, data, to_string):
        return schema.dumps(data).data if to_string else schema.dump(data).data


class BrewmasterSchemaParser(SchemaParser):
    def __init__(self):
        warnings.warn(
            "Reference made to 'BrewmasterSchemaParser'. This name will be removed in "
            "version 3.0, please use 'SchemaParser' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        super(BrewmasterSchemaParser, self).__init__()



# -*- coding: utf-8 -*-

import warnings
from argparse import ArgumentParser

from yapconf import YapconfSpec
from yapconf.exceptions import YapconfItemNotFound

from brewtils.decorators import command, parameter, system
from brewtils.errors import ValidationError
from brewtils.log import configure_logging
from brewtils.plugin import Plugin, RemotePlugin
from brewtils.rest import normalize_url_prefix
from brewtils.rest.easy_client import EasyClient
from brewtils.rest.system_client import SystemClient
from ._version import __version__ as generated_version
from .specification import SPECIFICATION

__all__ = [
    "command",
    "parameter",
    "system",
    "Plugin",
    "RemotePlugin",
    "EasyClient",
    "SystemClient",
    "get_easy_client",
    "get_argument_parser",
    "get_connection_info",
    "load_config",
    "get_bg_connection_parameters",
    "configure_logging",
]

__version__ = generated_version


def get_easy_client(**kwargs):
    """Easy way to get an EasyClient

    The benefit to this method over creating an EasyClient directly is that
    this method will also search the environment for parameters. Kwargs passed
    to this method will take priority, however.

    Args:
        **kwargs: Options for configuring the EasyClient

    Returns:
        :obj:`brewtils.rest.easy_client.EasyClient`: The configured client
    """
    from brewtils.rest.easy_client import EasyClient

    parser = kwargs.pop("parser", None)
    logger = kwargs.pop("logger", None)

    return EasyClient(logger=logger, parser=parser, **get_connection_info(**kwargs))


def get_argument_parser():
    """Get an ArgumentParser pre-populated with Brewtils arguments

    This is helpful if you're expecting additional command line arguments to
    a plugin startup script.

    This enables doing something like::

        def main():
            parser = get_argument_parser()
            parser.add_argument('positional_arg')

            parsed_args = parser.parse_args(sys.argv[1:])

            # Now you can use the extra argument
            client = MyClient(parsed_args.positional_arg)

            # But you'll need to be careful when using the 'normal' Brewtils
            # configuration loading methods:

            # Option 1: Tell Brewtils about your customized parser
            connection = get_connection_info(cli_args=sys.argv[1:],
                                             argument_parser=parser)

            # Option 2: Use the parsed CLI as a dictionary
            connection = get_connection_info(**vars(parsed_args))

            # Now specify connection kwargs like normal
            plugin = RemotePlugin(client, name=...
                                  **connection)

    IMPORTANT: Note that in both cases the returned ``connection`` object **will
    not** contain your new value. Both options just prevent normal CLI parsing
    from failing on the unknown argument.

    Returns:
        :ArgumentParser: Argument parser with Brewtils arguments loaded
    """
    parser = ArgumentParser()

    YapconfSpec(SPECIFICATION).add_arguments(parser)

    return parser


def get_connection_info(cli_args=None, argument_parser=None, **kwargs):
    """Wrapper around ``load_config`` that returns only connection parameters

    Args:
        cli_args (list, optional): List of command line arguments for
            configuration loading
        argument_parser (ArgumentParser, optional): Argument parser to use when
            parsing cli_args. Supplying this allows adding additional arguments
            prior to loading the configuration. This can be useful if your
            startup script takes additional arguments.
        **kwargs: Additional configuration overrides

    Returns:
        :dict: Parameters needed to make a connection to Beergarden
    """
    config = load_config(cli_args=cli_args, argument_parser=argument_parser, **kwargs)

    return {
        key: config[key]
        for key in (
            "bg_host",
            "bg_port",
            "ssl_enabled",
            "api_version",
            "ca_cert",
            "client_cert",
            "url_prefix",
            "ca_verify",
            "username",
            "password",
            "access_token",
            "refresh_token",
            "client_timeout",
        )
    }


def load_config(cli_args=None, argument_parser=None, **kwargs):
    """Load configuration using Yapconf

    Configuration will be loaded from these sources, with earlier sources having
    higher priority:

        1. ``**kwargs`` passed to this method
        2. ``cli_args`` passed to this method
        3. Environment variables using the ``BG_`` prefix
        4. Default values in the brewtils specification

    Args:
        cli_args (list, optional): List of command line arguments for
            configuration loading
        argument_parser (ArgumentParser, optional): Argument parser to use when
            parsing cli_args. Supplying this allows adding additional arguments
            prior to loading the configuration. This can be useful if your
            startup script takes additional arguments. See get_argument_parser
            for additional information.
        **kwargs: Additional configuration overrides

    Returns:
        :obj:`box.Box`: The resolved configuration object
    """
    spec = YapconfSpec(SPECIFICATION, env_prefix="BG_")

    sources = []

    if kwargs:
        # Do a little kwarg massaging for backwards compatibility
        if "bg_host" not in kwargs and "host" in kwargs:
            warnings.warn(
                "brewtils.load_config called with 'host' keyword "
                "argument. This name will be removed in version 3.0, "
                "please use 'bg_host' instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            kwargs["bg_host"] = kwargs.pop("host")
        if "bg_port" not in kwargs and "port" in kwargs:
            warnings.warn(
                "brewtils.load_config called with 'port' keyword "
                "argument. This name will be removed in version 3.0, "
                "please use 'bg_port' instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            kwargs["bg_port"] = kwargs.pop("port")

        sources.append(("kwargs", kwargs))

    if cli_args:
        if not argument_parser:
            argument_parser = ArgumentParser()
            spec.add_arguments(argument_parser)

        parsed_args = argument_parser.parse_args(cli_args)
        sources.append(("cli_args", vars(parsed_args)))

    sources.append("ENVIRONMENT")

    try:
        config = spec.load_config(*sources)
    except YapconfItemNotFound as ex:
        if ex.item.name == "bg_host":
            raise ValidationError(
                "Unable to create a plugin without a "
                "beer-garden host. Please specify one on the "
                "command line (--bg-host), in the "
                "environment (BG_HOST), or in kwargs "
                "(bg_host)"
            )
        raise

    # Make sure the url_prefix is normal
    config.url_prefix = normalize_url_prefix(config.url_prefix)

    return config


# Alias old names for compatibility
get_bg_connection_parameters = get_connection_info



# -*- coding: utf-8 -*-
import abc
import logging

import six
import threading
from functools import partial
from pika import BlockingConnection, URLParameters, BasicProperties, SelectConnection
from pika.exceptions import AMQPConnectionError

from brewtils.errors import DiscardMessageException, RepublishRequestException
from brewtils.queues import PikaClient, PIKA_ONE
from brewtils.schema_parser import SchemaParser

if PIKA_ONE:
    from pika.exceptions import (
        ConnectionClosed,
        ChannelClosedByBroker,
        ChannelClosedByClient,
    )


@six.add_metaclass(abc.ABCMeta)
class RequestConsumerBase(threading.Thread):
    """RabbitMQ message consumer

    This consumer is designed to be fault-tolerant - if RabbitMQ closes the
    connection the consumer will attempt to reopen it. There are limited
    reasons why the connection may be closed from the broker side and usually
    indicates permission related issues or socket timeouts.

    Unexpected channel closures can indicate a problem with a command that was
    issued.

    :param str amqp_url: The AMQP url to connection with
    :param str queue_name: The name of the queue to connect to
    :param func on_message_callback: The function called to invoke message
        processing. Must return a Future.
    :param event panic_event: An event to be set in the event of a catastrophic
        failure
    :type event: :py:class:`threading.Event`
    :param logger: A configured logger
    :type logger: :py:class:`logging.Logger`
    :param str thread_name: The name to use for this thread
    :param int max_connect_retries: Number of connection retry attempts before
        failure. Default is -1 (retry forever).
    :param int max_connect_backoff: Maximum amount of time to wait between
        connection retry attempts. Default 30.
    :param int max_concurrent: Maximum requests to process concurrently
    """

    def __init__(
        self,
        amqp_url=None,
        queue_name=None,
        on_message_callback=None,
        panic_event=None,
        logger=None,
        thread_name=None,
        **kwargs
    ):
        self._connection = None
        self._channel = None
        self._consumer_tag = None

        self._queue_name = queue_name
        self._on_message_callback = on_message_callback
        self._panic_event = panic_event
        self._max_connect_retries = kwargs.get("max_connect_retries", -1)
        self._max_connect_backoff = kwargs.get("max_connect_backoff", 30)
        self._max_concurrent = kwargs.get("max_concurrent", 1)
        self.logger = logger or logging.getLogger(__name__)
        self.shutdown_event = threading.Event()

        if kwargs.get("connection_info", None):
            pika_base = PikaClient(**kwargs["connection_info"])
            self._connection_parameters = pika_base.connection_parameters()
        else:
            self._connection_parameters = URLParameters(amqp_url)

        super(RequestConsumerBase, self).__init__(name=thread_name)

    def run(self):
        """Run the consumer

        Creates a connection to RabbitMQ and starts the IOLoop. The IOLoop will
        block and allow the SelectConnection to operate.

        :return:
        """
        self._connection = self.open_connection()

        # It is possible to return from open_connection without acquiring a
        # connection. This usually happens if no max_connect_retries was set
        # and we are constantly trying to connect to a queue that does not
        # exist. For those cases, there is no reason to start an ioloop.
        if self._connection:
            self._connection.ioloop.start()

    def stop(self):
        """Cleanly shutdown the connection

        Assumes the stop_consuming method has already been called. When the
        queueing service acknowledges the closure, the connection is closed
        which will end the RequestConsumer.

        :return:
        """
        self.logger.debug("Stopping request consumer")
        self.shutdown_event.set()
        self.close_channel()

    def on_message(self, channel, basic_deliver, properties, body):
        """Invoked when a message is delivered from the queueing service

        Invoked by pika when a message is delivered from RabbitMQ. The channel
        is passed for your convenience. The basic_deliver object that is passed
        in carries the exchange, routing key, delivery tag and a redelivered
        flag for the message. the properties passed in is an instance of
        BasicProperties with the message properties and the body is the message
        that was sent.

        :param pika.channel.Channel channel: The channel object
        :param pika.Spec.Basic.Deliver basic_deliver: basic_deliver method
        :param pika.Spec.BasicProperties properties: properties
        :param bytes body: The message body
        """
        self.logger.debug(
            "Received message #%s from %s on channel %s: %s",
            basic_deliver.delivery_tag,
            properties.app_id,
            channel.channel_number,
            body,
        )

        # Pika gives us bytes, but we want a string to be ok too
        try:
            body = body.decode()
        except AttributeError:
            pass

        try:
            future = self._on_message_callback(body, properties.headers)
            callback = partial(self.on_message_callback_complete, basic_deliver)
            future.add_done_callback(callback)
        except DiscardMessageException:
            self.logger.debug(
                "Nacking message %s, not attempting to requeue",
                basic_deliver.delivery_tag,
            )
            self._channel.basic_nack(basic_deliver.delivery_tag, requeue=False)
        except Exception as ex:
            self.logger.exception(
                "Exception while trying to schedule message %s, about to nack "
                "and requeue: %s",
                basic_deliver.delivery_tag,
                ex,
            )
            self._channel.basic_nack(basic_deliver.delivery_tag, requeue=True)

    def on_message_callback_complete(self, basic_deliver, future):
        """Invoked when the future returned by _on_message_callback completes.

        :param pika.Spec.Basic.Deliver basic_deliver: basic_deliver method
        :param concurrent.futures.Future future: Completed future
        :return: None
        """
        delivery_tag = basic_deliver.delivery_tag

        if not future.exception():
            try:
                self.logger.debug("Acking message %s", delivery_tag)
                self._channel.basic_ack(delivery_tag)
            except Exception as ex:
                self.logger.exception(
                    "Error acking message %s, about to shut down: %s", delivery_tag, ex
                )
                self._panic_event.set()
        else:
            real_ex = future.exception()

            if isinstance(real_ex, RepublishRequestException):
                try:
                    with BlockingConnection(self._connection_parameters) as c:
                        headers = real_ex.headers
                        headers.update({"request_id": real_ex.request.id})
                        props = BasicProperties(
                            app_id="beer-garden",
                            content_type="text/plain",
                            headers=headers,
                            priority=1,
                        )
                        c.channel().basic_publish(
                            exchange=basic_deliver.exchange,
                            properties=props,
                            routing_key=basic_deliver.routing_key,
                            body=SchemaParser.serialize_request(real_ex.request),
                        )

                    self._channel.basic_ack(delivery_tag)
                except Exception as ex:
                    self.logger.exception(
                        "Error republishing message %s, about to shut down: %s",
                        delivery_tag,
                        ex,
                    )
                    self._panic_event.set()
            elif isinstance(real_ex, DiscardMessageException):
                self.logger.info(
                    "Nacking message %s, not attempting to requeue", delivery_tag
                )
                self._channel.basic_nack(delivery_tag, requeue=False)
            else:
                # If request processing throws anything else we terminate
                self.logger.exception(
                    "Unexpected exception during request %s processing, about "
                    "to shut down: %s",
                    delivery_tag,
                    real_ex,
                    exc_info=False,
                )
                self._panic_event.set()

    def open_connection(self):
        """Opens a connection to RabbitMQ

        This method connects to RabbitMQ, returning the connection handle. The
        on_connection_open method will be invoked when the connection opens.

        :rtype: pika.SelectConnection
        """
        time_to_wait = 0.1
        retries = 0
        while not self.shutdown_event.is_set():
            try:
                return SelectConnection(
                    self._connection_parameters,
                    self.on_connection_open,
                    **self._select_kwargs()
                )
            except AMQPConnectionError as ex:
                if 0 <= self._max_connect_retries <= retries:
                    raise ex
                self.logger.warning(
                    "Error attempting to connect, waiting %s seconds and "
                    "attempting again" % time_to_wait
                )
                self.shutdown_event.wait(time_to_wait)
                time_to_wait = min(time_to_wait * 2, self._max_connect_backoff)
                retries += 1

    def on_connection_open(self, unused_connection):
        """Invoked when the connection has been established

        This method is called by pika once the connection to RabbitMQ has been
        established. It passes the handle to the connection object in case we
        need it, but in this case, we'll just mark it unused.

        :type unused_connection: pika.SelectConnection
        """
        self.logger.debug("Connection opened: %s", unused_connection)
        self._connection.add_on_close_callback(self.on_connection_closed)
        self.open_channel()

    def close_connection(self):
        """This method closes the connection to RabbitMQ"""
        self.logger.debug("Closing connection")
        self._connection.close()

    def do_on_connection_closed(self, connection, reply_code, reply_text):
        """Invoked when the connection is closed

        This method is invoked by pika when the connection to RabbitMQ is closed
        unexpectedly. This method will attempt to reconnect.

        :param pika.connection.Connection connection: the closed connection
        :param int reply_code: The server provided reply_code if given
        :param basestring reply_text: The server provided reply_text if given
        """
        self.logger.debug(
            'Connection "%s" closed: (%s) %s' % (connection, reply_code, reply_text)
        )
        self._channel = None

        # A 320 is the server forcing the connection to close
        if reply_code == 320:
            self.shutdown_event.set()

        if self.shutdown_event.is_set():
            self._connection.ioloop.stop()
        else:
            self.logger.warning(
                "Connection unexpectedly closed: (%s) %s" % (reply_code, reply_text)
            )
            self.logger.warning("Attempting to reopen connection in 5 seconds")
            self._connection.add_timeout(5, self.reconnect)

    def reconnect(self):
        """Will be invoked by the IOLoop timer if the connection is closed"""

        # This is the old connection IOLoop instance, stop its ioloop
        self._connection.ioloop.stop()

        if not self.shutdown_event.is_set():
            # Creates a new connection
            self._connection = self.open_connection()

            # There is now a new connection, needs a new ioloop to run
            if self._connection:
                self._connection.ioloop.start()

    def open_channel(self):
        """Open a channel using the connection

        When RabbitMQ responds that the channel is open, the on_channel_open
        callback will be invoked.
        """
        self.logger.debug("Opening a new channel")
        self._connection.channel(on_open_callback=self.on_channel_open)

    def on_channel_open(self, channel):
        """Invoked when the channel has been opened

        Immediately start consuming since the queue bindings are not the
        consumer's responsibility.

        :param pika.channel.Channel channel: The channel object
        """
        self.logger.debug("Channel opened: %s", channel)
        self._channel = channel
        self._channel.add_on_close_callback(self.on_channel_closed)
        self.start_consuming()

    def close_channel(self):
        """Cleanly close the channel"""
        self.logger.debug("Closing the channel")
        self._channel.close()

    def do_on_channel_closed(self, channel, reply_code, reply_text):
        """Invoked when the connection is closed

        Invoked by pika when RabbitMQ unexpectedly closes the channel. Channels
        are usually closed as a result of something that violates the protocol,
        such as attempting to re-declare an exchange or queue with different
        parameters.

        This indicates that something has gone wrong, so close the connection
        to reset.

        :param pika.channel.Channel channel: The closed channel
        :param int reply_code: The numeric reason the channel was closed
        :param str reply_text: The text reason the channel was closed
        """
        self.logger.debug(
            "Channel %i was closed: (%s) %s" % (int(channel), reply_code, reply_text)
        )
        self._connection.close()

    def start_consuming(self):
        """Begin consuming messages

        The RabbitMQ prefetch is set to the maximum number of concurrent
        consumers. This ensures that messages remain in RabbitMQ until a
        consuming thread is available to process them.

        An on_cancel_callback is registered so that the consumer is notified if
        it is canceled by the broker.
        """
        self.logger.debug("Issuing consumer related RPC commands")

        self._channel.basic_qos(prefetch_count=self._max_concurrent)
        self._channel.add_on_cancel_callback(self.on_consumer_cancelled)
        self._consumer_tag = self._channel.basic_consume(**self._consume_kwargs())

    def stop_consuming(self):
        """Stop consuming messages"""
        self.logger.debug("Stopping consuming on channel %s", self._channel)
        if self._channel:
            self.logger.debug("Sending a Basic.Cancel RPC command to RabbitMQ")
            self._channel.basic_cancel(
                callback=self.on_cancelok, consumer_tag=self._consumer_tag
            )

    def on_consumer_cancelled(self, method_frame):
        """Invoked when the consumer is canceled by the broker

        This method will simply close the channel if it exists.

        :param pika.frame.Method method_frame: The Basic.Cancel frame
        """
        self.logger.debug(
            "Consumer was cancelled remotely, shutting down: %r" % method_frame
        )
        if self._channel:
            self.close_channel()

    def on_cancelok(self, unused_frame):
        """Invoked when RabbitMq acknowledges consumer cancellation

        This method is invoked when RabbitMQ acknowledges the cancellation of a
        consumer. It is unused except for logging purposes.

        :param pika.frame.Method unused_frame: The Basic.CancelOK frame
        """
        self.logger.debug(unused_frame)
        self.logger.debug("RabbitMQ acknowledged consumer cancellation")


class RequestConsumerPika0(RequestConsumerBase):
    """Implementation of a Pika v0 RequestConsumer

    This exists because some kwargs and callback signatures changed between version
    0 and version 1. This is essentially a wrapper that delegates to the
    RequestConsumerBase methods.

    """

    def on_connection_closed(self, *args):
        self.do_on_connection_closed(*args)

    def on_channel_closed(self, *args):
        self.do_on_channel_closed(*args)

    @staticmethod
    def _select_kwargs():
        return {"stop_ioloop_on_close": False}

    def _consume_kwargs(self):
        return {"queue": self._queue_name, "consumer_callback": self.on_message}


class RequestConsumerPika1(RequestConsumerBase):
    """Implementation of a Pika v1 RequestConsumer

    This exists because some kwargs and callback signatures changed between version
    0 and version 1. This is essentially a wrapper that delegates to the
    RequestConsumerBase methods after translating arguments.

    """

    def on_connection_closed(self, connection, exc):
        if isinstance(exc, ConnectionClosed):
            self.do_on_connection_closed(connection, exc.reply_code, exc.reply_text)
        else:
            raise exc

    def on_channel_closed(self, channel, exc):
        if isinstance(exc, (ChannelClosedByBroker, ChannelClosedByClient)):
            self.do_on_channel_closed(channel, exc.reply_code, exc.reply_text)
        else:
            raise exc

    @staticmethod
    def _select_kwargs():
        return {}

    def _consume_kwargs(self):
        return {"queue": self._queue_name, "on_message_callback": self.on_message}


# The real RequestConsumer is based on the pika version
RequestConsumer = RequestConsumerPika1 if PIKA_ONE else RequestConsumerPika0



# -*- coding: utf-8 -*-

from enum import Enum

import six

from brewtils.errors import RequestStatusTransitionError

__all__ = [
    "System",
    "Instance",
    "Command",
    "Parameter",
    "Request",
    "PatchOperation",
    "Choices",
    "LoggingConfig",
    "Event",
    "Events",
    "Queue",
    "Principal",
    "Role",
    "RefreshToken",
    "Job",
    "RequestTemplate",
    "DateTrigger",
    "CronTrigger",
    "IntervalTrigger",
]


class Events(Enum):
    BREWVIEW_STARTED = 1
    BREWVIEW_STOPPED = 2
    BARTENDER_STARTED = 3
    BARTENDER_STOPPED = 4
    REQUEST_CREATED = 5
    REQUEST_STARTED = 6
    REQUEST_COMPLETED = 7
    INSTANCE_INITIALIZED = 8
    INSTANCE_STARTED = 9
    INSTANCE_STOPPED = 10
    SYSTEM_CREATED = 11
    SYSTEM_UPDATED = 12
    SYSTEM_REMOVED = 13
    QUEUE_CLEARED = 14
    ALL_QUEUES_CLEARED = 15


class Command(object):

    schema = "CommandSchema"

    COMMAND_TYPES = ("ACTION", "INFO", "EPHEMERAL")
    OUTPUT_TYPES = ("STRING", "JSON", "XML", "HTML")

    def __init__(
        self,
        name=None,
        description=None,
        id=None,
        parameters=None,
        command_type=None,
        output_type=None,
        schema=None,
        form=None,
        template=None,
        icon_name=None,
        system=None,
    ):
        self.name = name
        self.description = description
        self.id = id
        self.parameters = parameters or []
        self.command_type = command_type
        self.output_type = output_type
        self.schema = schema
        self.form = form
        self.template = template
        self.icon_name = icon_name
        self.system = system

    def __str__(self):
        return self.name

    def __repr__(self):
        return "<Command: %s>" % self.name

    def parameter_keys(self):
        """Convenience Method for returning all the keys of this command's parameters.

        :return list_of_parameters:
        """
        return [p.key for p in self.parameters]

    def get_parameter_by_key(self, key):
        """Given a Key, it will return the parameter (or None) with that key

        :param key:
        :return parameter:
        """
        for parameter in self.parameters:
            if parameter.key == key:
                return parameter

        return None

    def has_different_parameters(self, parameters):
        """Given a set of parameters, determines if the parameters provided differ from the
        parameters already defined on this command.

        :param parameters:
        :return boolean:
        """
        if len(parameters) != len(self.parameters):
            return True

        for parameter in parameters:
            if parameter.key not in self.parameter_keys():
                return True

            current_param = self.get_parameter_by_key(parameter.key)
            if current_param.is_different(parameter):
                return True

        return False


class Instance(object):

    schema = "InstanceSchema"

    INSTANCE_STATUSES = {
        "INITIALIZING",
        "RUNNING",
        "PAUSED",
        "STOPPED",
        "DEAD",
        "UNRESPONSIVE",
        "STARTING",
        "STOPPING",
        "UNKNOWN",
    }

    def __init__(
        self,
        name=None,
        description=None,
        id=None,
        status=None,
        status_info=None,
        queue_type=None,
        queue_info=None,
        icon_name=None,
        metadata=None,
    ):
        self.name = name
        self.description = description
        self.id = id
        self.status = status.upper() if status else None
        self.status_info = status_info or {}
        self.queue_type = queue_type
        self.queue_info = queue_info or {}
        self.icon_name = icon_name
        self.metadata = metadata or {}

    def __str__(self):
        return self.name

    def __repr__(self):
        return "<Instance: name=%s, status=%s>" % (self.name, self.status)


class Choices(object):

    schema = "ChoicesSchema"

    TYPES = ("static", "url", "command")
    DISPLAYS = ("select", "typeahead")

    def __init__(self, type=None, display=None, value=None, strict=None, details=None):
        self.type = type
        self.strict = strict
        self.value = value
        self.display = display
        self.details = details or {}

    def __str__(self):
        return self.value.__str__()

    def __repr__(self):
        return "<Choices: type=%s, display=%s, value=%s>" % (
            self.type,
            self.display,
            self.value,
        )


class Parameter(object):

    schema = "ParameterSchema"

    TYPES = (
        "String",
        "Integer",
        "Float",
        "Boolean",
        "Any",
        "Dictionary",
        "Date",
        "DateTime",
    )
    FORM_INPUT_TYPES = ("textarea",)

    def __init__(
        self,
        key,
        type=None,
        multi=None,
        display_name=None,
        optional=None,
        default=None,
        description=None,
        choices=None,
        parameters=None,
        nullable=None,
        maximum=None,
        minimum=None,
        regex=None,
        form_input_type=None,
    ):
        self.key = key
        self.type = type
        self.multi = multi
        self.display_name = display_name
        self.optional = optional
        self.default = default
        self.description = description
        self.choices = choices
        self.parameters = parameters or []
        self.nullable = nullable
        self.maximum = maximum
        self.minimum = minimum
        self.regex = regex
        self.form_input_type = form_input_type

    def __str__(self):
        return self.key

    def __repr__(self):
        return "<Parameter: key=%s, type=%s, description=%s>" % (
            self.key,
            self.type,
            self.description,
        )

    def is_different(self, other):
        if not type(other) is type(self):
            return True

        fields_to_compare = [
            "key",
            "type",
            "multi",
            "optional",
            "default",
            "nullable",
            "maximum",
            "minimum",
            "regex",
        ]
        for field in fields_to_compare:
            if getattr(self, field) != getattr(other, field):
                return True

        if len(self.parameters) != len(other.parameters):
            return True

        parameter_keys = [p.key for p in self.parameters]
        for parameter in other.parameters:
            if parameter.key not in parameter_keys:
                return True

            current_param = list(
                filter((lambda p: p.key == parameter.key), self.parameters)
            )[0]
            if current_param.is_different(parameter):
                return True

        return False


class RequestTemplate(object):

    schema = "RequestTemplateSchema"

    def __init__(
        self,
        system=None,
        system_version=None,
        instance_name=None,
        command=None,
        parameters=None,
        comment=None,
        metadata=None,
    ):
        self.system = system
        self.system_version = system_version
        self.instance_name = instance_name
        self.command = command
        self.parameters = parameters
        self.comment = comment
        self.metadata = metadata or {}

    def __str__(self):
        return self.command

    def __repr__(self):
        return (
            "<RequestTemplate: command=%s, system=%s, system_version=%s, "
            "instance_name=%s>"
            % (self.command, self.system, self.system_version, self.instance_name)
        )


class Request(RequestTemplate):

    schema = "RequestSchema"

    STATUS_LIST = ("CREATED", "RECEIVED", "IN_PROGRESS", "CANCELED", "SUCCESS", "ERROR")
    COMPLETED_STATUSES = ("CANCELED", "SUCCESS", "ERROR")
    COMMAND_TYPES = ("ACTION", "INFO", "EPHEMERAL")
    OUTPUT_TYPES = ("STRING", "JSON", "XML", "HTML")

    def __init__(
        self,
        system=None,
        system_version=None,
        instance_name=None,
        command=None,
        id=None,
        parent=None,
        children=None,
        parameters=None,
        comment=None,
        output=None,
        output_type=None,
        status=None,
        command_type=None,
        created_at=None,
        error_class=None,
        metadata=None,
        updated_at=None,
        has_parent=None,
        requester=None,
    ):
        super(Request, self).__init__(
            system=system,
            system_version=system_version,
            instance_name=instance_name,
            command=command,
            parameters=parameters,
            comment=comment,
            metadata=metadata,
        )
        self.id = id
        self.parent = parent
        self.children = children
        self.output = output
        self.output_type = output_type
        self._status = status
        self.command_type = command_type
        self.created_at = created_at
        self.updated_at = updated_at
        self.error_class = error_class
        self.has_parent = has_parent
        self.requester = requester

    def __repr__(self):
        return (
            "<Request: command=%s, status=%s, system=%s, system_version=%s, "
            "instance_name=%s>"
            % (
                self.command,
                self.status,
                self.system,
                self.system_version,
                self.instance_name,
            )
        )

    @property
    def status(self):
        return self._status

    @property
    def is_ephemeral(self):
        return self.command_type and self.command_type.upper() == "EPHEMERAL"

    @property
    def is_json(self):
        return self.output_type and self.output_type.upper() == "JSON"

    @status.setter
    def status(self, value):
        if self._status in self.COMPLETED_STATUSES:
            raise RequestStatusTransitionError(
                "Status for a request cannot be updated once it has been "
                "completed. Current status: {0} Requested status: {1}".format(
                    self.status, value
                )
            )

        elif self._status == "IN_PROGRESS" and value not in self.COMPLETED_STATUSES + (
            "IN_PROGRESS",
        ):
            raise RequestStatusTransitionError(
                "A request cannot go from IN_PROGRESS to a non-completed "
                "status. Completed statuses are {0}. You requested: {1}".format(
                    self.COMPLETED_STATUSES, value
                )
            )

        self._status = value


class System(object):

    schema = "SystemSchema"

    def __init__(
        self,
        name=None,
        description=None,
        version=None,
        id=None,
        max_instances=None,
        instances=None,
        commands=None,
        icon_name=None,
        display_name=None,
        metadata=None,
    ):
        self.name = name
        self.description = description
        self.version = version
        self.id = id
        self.max_instances = max_instances
        self.instances = instances or []
        self.commands = commands or []
        self.icon_name = icon_name
        self.display_name = display_name
        self.metadata = metadata or {}

    def __str__(self):
        return "%s-%s" % (self.name, self.version)

    def __repr__(self):
        return "<System: name=%s, version=%s>" % (self.name, self.version)

    @property
    def instance_names(self):
        return [i.name for i in self.instances]

    def has_instance(self, name):
        """Determine if an instance currently exists in the system

        :param name: The name of the instance to search
        :return: True if an instance with the given name exists for this system, False otherwise.
        """
        return True if self.get_instance(name) else False

    def get_instance(self, name):
        """Get an instance that currently exists in the system

        :param name: The name of the instance to search
        :return: The instance with the given name exists for this system, None otherwise
        """
        for instance in self.instances:
            if instance.name == name:
                return instance
        return None

    def get_command_by_name(self, command_name):
        """Retrieve a particular command from the system

        :param command_name: Name of the command to retrieve
        :return: The command object. None if the given command name does not exist in this system.
        """
        for command in self.commands:
            if command.name == command_name:
                return command

        return None

    def has_different_commands(self, commands):
        """Check if a set of commands is different than the current commands

        :param commands: The set commands to compare against the current set
        :return: True if the sets are different, False if the sets are the same
        """
        if len(commands) != len(self.commands):
            return True

        for command in commands:
            if command.name not in [c.name for c in self.commands]:
                return True

            current_command = self.get_command_by_name(command.name)

            if current_command.has_different_parameters(command.parameters):
                return True

        return False


class PatchOperation(object):

    schema = "PatchSchema"

    def __init__(self, operation=None, path=None, value=None):
        self.operation = operation
        self.path = path
        self.value = value

    def __str__(self):
        return "%s, %s, %s" % (self.operation, self.path, self.value)

    def __repr__(self):
        return "<Patch: operation=%s, path=%s, value=%s>" % (
            self.operation,
            self.path,
            self.value,
        )


class LoggingConfig(object):

    schema = "LoggingConfigSchema"

    LEVELS = ("DEBUG", "INFO", "WARN", "ERROR")
    SUPPORTED_HANDLERS = ("stdout", "file", "logstash")

    DEFAULT_FORMAT = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    DEFAULT_HANDLER = {
        "class": "logging.StreamHandler",
        "stream": "ext::/sys.stdout",
        "formatter": "default",
    }

    def __init__(self, level=None, handlers=None, formatters=None, loggers=None):
        self.level = level
        self.handlers = handlers
        self.formatters = formatters
        self._loggers = loggers or {}

    @property
    def handler_names(self):
        if self.handlers:
            return set(self.handlers)
        else:
            return None

    @property
    def formatter_names(self):
        if self.formatters:
            return set(self.formatters)
        else:
            return None

    def get_plugin_log_config(self, **kwargs):
        """Get a specific plugin logging configuration.

        It is possible for different systems to have different logging configurations. This method
        will create the correct plugin logging configuration and return it. If a specific logger
        is not found for a system, then the current logging configuration will be returned.

        :param kwargs: Identifying information for a system (i.e. system_name)
        :return:
        """
        system_name = kwargs.pop("system_name", None)
        specific_logger = self._loggers.get(system_name, {})

        # If there is no specific logger, then we simply return this object
        # otherwise, we need to construct a new LoggingConfig object with
        # the overrides given in the logger.
        if not specific_logger:
            return self

        level = specific_logger.get("level", self.level)
        handlers = self._generate_handlers(specific_logger.get("handlers"))
        formatters = self._generate_formatters(specific_logger.get("formatters", {}))

        return LoggingConfig(level=level, handlers=handlers, formatters=formatters)

    def _generate_handlers(self, specific_handlers):

        # If we are not given an override for handlers, then we will just
        # assume that we want to use all the handlers given in the current
        # configuration.
        if not specific_handlers:
            return self.handlers

        if isinstance(specific_handlers, list):
            handlers = {}
            for handler_name in specific_handlers:
                handlers[handler_name] = self.handlers[handler_name]
        else:
            return specific_handlers

        return handlers

    def _generate_formatters(self, specific_formatters):

        # If we are not given an override for formatters, then we will just
        # assume that we want to use the formatters given in the current
        # configuration
        if not specific_formatters:
            return self.formatters

        # In case no formatter is provided, we always want a default.
        formatters = {"default": {"format": self.DEFAULT_FORMAT}}
        for formatter_name, format_str in six.iteritems(specific_formatters):
            formatters[formatter_name] = {"format": format_str}

        return formatters

    def __str__(self):
        return "%s, %s, %s" % (self.level, self.handler_names, self.formatter_names)

    def __repr__(self):
        return "<LoggingConfig: level=%s, handlers=%s, formatters=%s" % (
            self.level,
            self.handler_names,
            self.formatter_names,
        )


class Event(object):

    schema = "EventSchema"

    def __init__(
        self, name=None, payload=None, error=None, metadata=None, timestamp=None
    ):
        self.name = name
        self.payload = payload
        self.error = error
        self.metadata = metadata or {}
        self.timestamp = timestamp

    def __str__(self):
        return "%s: %s, %s" % (self.name, self.payload, self.metadata)

    def __repr__(self):
        return "<Event: name=%s, error=%s, payload=%s, metadata=%s>" % (
            self.name,
            self.error,
            self.payload,
            self.metadata,
        )


class Queue(object):

    schema = "QueueSchema"

    def __init__(
        self,
        name=None,
        system=None,
        version=None,
        instance=None,
        system_id=None,
        display=None,
        size=None,
    ):
        self.name = name
        self.system = system
        self.version = version
        self.instance = instance
        self.system_id = system_id
        self.display = display
        self.size = size

    def __str__(self):
        return "%s: %s" % (self.name, self.size)

    def __repr__(self):
        return "<Queue: name=%s, size=%s>" % (self.name, self.size)


class Principal(object):

    schema = "PrincipalSchema"

    def __init__(
        self,
        id=None,
        username=None,
        roles=None,
        permissions=None,
        preferences=None,
        metadata=None,
    ):
        self.id = id
        self.username = username
        self.roles = roles
        self.permissions = permissions
        self.preferences = preferences
        self.metadata = metadata

    def __str__(self):
        return "%s" % self.username

    def __repr__(self):
        return "<Principal: username=%s, roles=%s, permissions=%s>" % (
            self.username,
            self.roles,
            self.permissions,
        )


class Role(object):

    schema = "RoleSchema"

    def __init__(
        self, id=None, name=None, description=None, roles=None, permissions=None
    ):
        self.id = id
        self.name = name
        self.description = description
        self.roles = roles
        self.permissions = permissions

    def __str__(self):
        return "%s" % self.name

    def __repr__(self):
        return "<Role: name=%s, roles=%s, permissions=%s>" % (
            self.name,
            self.roles,
            self.permissions,
        )


class RefreshToken(object):

    schema = "RefreshTokenSchema"

    def __init__(self, id=None, issued=None, expires=None, payload=None):
        self.id = id
        self.issued = issued
        self.expires = expires
        self.payload = payload or {}

    def __str__(self):
        return "%s" % self.payload

    def __repr__(self):
        return "<RefreshToken: issued=%s, expires=%s, payload=%s>" % (
            self.issued,
            self.expires,
            self.payload,
        )


class Job(object):

    TRIGGER_TYPES = {"interval", "date", "cron"}
    STATUS_TYPES = {"RUNNING", "PAUSED"}
    schema = "JobSchema"

    def __init__(
        self,
        id=None,
        name=None,
        trigger_type=None,
        trigger=None,
        request_template=None,
        misfire_grace_time=None,
        coalesce=None,
        next_run_time=None,
        success_count=None,
        error_count=None,
        status=None,
        max_instances=None,
    ):
        self.id = id
        self.name = name
        self.trigger_type = trigger_type
        self.trigger = trigger
        self.request_template = request_template
        self.misfire_grace_time = misfire_grace_time
        self.coalesce = coalesce
        self.next_run_time = next_run_time
        self.success_count = success_count
        self.error_count = error_count
        self.status = status
        self.max_instances = max_instances

    def __str__(self):
        return "%s: %s" % (self.name, self.id)

    def __repr__(self):
        return "<Job: name=%s, id=%s>" % (self.name, self.id)


class DateTrigger(object):
    schema = "DateTriggerSchema"

    def __init__(self, run_date=None, timezone=None):
        self.run_date = run_date
        self.timezone = timezone

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return "<DateTrigger: run_date=%s>" % self.run_date


class IntervalTrigger(object):
    schema = "IntervalTriggerSchema"

    def __init__(
        self,
        weeks=None,
        days=None,
        hours=None,
        minutes=None,
        seconds=None,
        start_date=None,
        end_date=None,
        timezone=None,
        jitter=None,
        reschedule_on_finish=None,
    ):
        self.weeks = weeks
        self.days = days
        self.hours = hours
        self.minutes = minutes
        self.seconds = seconds
        self.start_date = start_date
        self.end_date = end_date
        self.timezone = timezone
        self.jitter = jitter
        self.reschedule_on_finish = reschedule_on_finish

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return (
            "<IntervalTrigger: weeks=%d, days=%d, hours=%d, "
            "minutes=%d, seconds=%d>"
            % (self.weeks, self.days, self.hours, self.minutes, self.seconds)
        )


class CronTrigger(object):
    schema = "CronTriggerSchema"

    def __init__(
        self,
        year=None,
        month=None,
        day=None,
        week=None,
        day_of_week=None,
        hour=None,
        minute=None,
        second=None,
        start_date=None,
        end_date=None,
        timezone=None,
        jitter=None,
    ):
        self.year = year
        self.month = month
        self.day = day
        self.week = week
        self.day_of_week = day_of_week
        self.hour = hour
        self.minute = minute
        self.second = second
        self.start_date = start_date
        self.end_date = end_date
        self.timezone = timezone
        self.jitter = jitter

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return "<CronTrigger: %s %s %s %s %s>" % (
            self.minute,
            self.hour,
            self.day,
            self.month,
            self.day,
        )



# -*- coding: utf-8 -*-

import json
import logging
import logging.config
import os
import sys
import threading
import warnings
from concurrent.futures import ThreadPoolExecutor

import six
from requests import ConnectionError as RequestsConnectionError

import brewtils
from brewtils.errors import (
    ValidationError,
    RequestProcessingError,
    DiscardMessageException,
    RepublishRequestException,
    RestConnectionError,
    PluginValidationError,
    RestClientError,
    parse_exception_as_json,
)
from brewtils.log import DEFAULT_LOGGING_CONFIG
from brewtils.models import Instance, Request, System
from brewtils.request_consumer import RequestConsumer
from brewtils.rest.easy_client import EasyClient
from brewtils.schema_parser import SchemaParser

request_context = threading.local()


class Plugin(object):
    """A beer-garden Plugin.

    This class represents a beer-garden Plugin - a continuously-running process
    that can receive and process Requests.

    To work, a Plugin needs a Client instance - an instance of a class defining
    which Requests this plugin can accept and process. The easiest way to define
     a ``Client`` is by annotating a class with the ``@system`` decorator.

    When creating a Plugin you can pass certain keyword arguments to let the
    Plugin know how to communicate with the beer-garden instance. These are:

        - ``bg_host``
        - ``bg_port``
        - ``ssl_enabled``
        - ``ca_cert``
        - ``client_cert``
        - ``bg_url_prefix``

    A Plugin also needs some identifying data. You can either pass parameters to
    the Plugin or pass a fully defined System object (but not both). Note that
    some fields are optional::

        Plugin(
            name="Test",
            version="1.0.0",
            instance_name="default",
            description="A Test",
        )

    or::

        the_system = System(
            name="Test",
            version="1.0.0",
            instance_name="default,
            description="A Test",
        )
        Plugin(system=the_system)

    If passing parameters directly note that these fields are required:

    name
        Environment variable ``BG_NAME`` will be used if not specified

    version
        Environment variable ``BG_VERSION`` will be used if not specified

    instance_name
        Environment variable ``BG_INSTANCE_NAME`` will be used if not specified.
        'default' will be used if not specified and loading from environment
        variable was unsuccessful

    And these fields are optional:

    - description  (Will use docstring summary line from Client if unspecified)
    - icon_name
    - metadata
    - display_name

    Plugins service requests using a
    :py:class:`concurrent.futures.ThreadPoolExecutor`. The maximum number of
    threads available is controlled by the max_concurrent argument (the
    'multithreaded' argument has been deprecated).

    .. warning::
        The default value for ``max_concurrent`` is 1. This means that a Plugin
        that invokes a Command on itself in the course of processing a Request
        will deadlock! If you intend to do this, please set ``max_concurrent``
        to a value that makes sense and be aware that Requests are processed in
        separate thread contexts!

    :param client: Instance of a class annotated with @system.
    :param str bg_host: Hostname of a beer-garden.
    :param int bg_port: Port beer-garden is listening on.
    :param bool ssl_enabled: Whether to use SSL for beer-garden communication.
    :param ca_cert: Certificate that issued the server certificate used by the
        beer-garden server.
    :param client_cert: Certificate used by the server making the connection to
        beer-garden.
    :param system: The system definition.
    :param name: The system name.
    :param description: The system description.
    :param version: The system version.
    :param icon_name: The system icon name.
    :param str instance_name: The name of the instance.
    :param logger: A logger that will be used by the Plugin.
    :type logger: :py:class:`logging.Logger`.
    :param parser: The parser to use when communicating with beer-garden.
    :type parser: :py:class:`brewtils.schema_parser.SchemaParser`.
    :param bool multithreaded: DEPRECATED Process requests in a separate thread.
    :param int worker_shutdown_timeout: Time to wait during shutdown to finish
        processing.
    :param dict metadata: Metadata specific to this plugin.
    :param int max_concurrent: Maximum number of requests to process
        concurrently.
    :param str bg_url_prefix: URL Prefix beer-garden is on.
    :param str display_name: The display name to use for the system.
    :param int max_attempts: Number of times to attempt updating the request
        before giving up (default -1 aka never).
    :param int max_timeout: Maximum amount of time to wait before retrying to
        update a request.
    :param int starting_timeout: Initial time to wait before the first retry.
    :param int max_instances: Max number of instances allowed for the system.
    :param bool ca_verify: Verify server certificate when making a request.
    :param str username: The username for Beergarden authentication
    :param str password: The password for Beergarden authentication
    :param access_token: Access token for Beergarden authentication
    :param refresh_token: Refresh token for Beergarden authentication
    """

    def __init__(
        self,
        client,
        bg_host=None,
        bg_port=None,
        ssl_enabled=None,
        ca_cert=None,
        client_cert=None,
        system=None,
        name=None,
        description=None,
        version=None,
        icon_name=None,
        instance_name=None,
        logger=None,
        parser=None,
        multithreaded=None,
        metadata=None,
        max_concurrent=None,
        bg_url_prefix=None,
        **kwargs
    ):
        # If a logger is specified or the logging module already has additional
        # handlers then we assume that logging has already been configured
        if logger or len(logging.getLogger(__name__).root.handlers) > 0:
            self.logger = logger or logging.getLogger(__name__)
            self._custom_logger = True
        else:
            logging.config.dictConfig(DEFAULT_LOGGING_CONFIG)
            self.logger = logging.getLogger(__name__)
            self._custom_logger = False

        connection_parameters = brewtils.get_connection_info(
            bg_host=bg_host,
            bg_port=bg_port,
            ssl_enabled=ssl_enabled,
            ca_cert=ca_cert,
            client_cert=client_cert,
            url_prefix=bg_url_prefix or kwargs.get("url_prefix", None),
            ca_verify=kwargs.get("ca_verify", None),
            username=kwargs.get("username", None),
            password=kwargs.get("password", None),
            client_timeout=kwargs.get("client_timeout", None),
        )
        self.bg_host = connection_parameters["bg_host"]
        self.bg_port = connection_parameters["bg_port"]
        self.ssl_enabled = connection_parameters["ssl_enabled"]
        self.ca_cert = connection_parameters["ca_cert"]
        self.client_cert = connection_parameters["client_cert"]
        self.bg_url_prefix = connection_parameters["url_prefix"]
        self.ca_verify = connection_parameters["ca_verify"]

        self.max_attempts = kwargs.get("max_attempts", -1)
        self.max_timeout = kwargs.get("max_timeout", 30)
        self.starting_timeout = kwargs.get("starting_timeout", 5)

        self.max_concurrent = self._setup_max_concurrent(multithreaded, max_concurrent)
        self.instance_name = instance_name or os.environ.get(
            "BG_INSTANCE_NAME", "default"
        )
        self.metadata = metadata or {}

        self.instance = None
        self.queue_connection_params = None
        self.admin_consumer = None
        self.request_consumer = None
        self.connection_poll_thread = None
        self.client = client
        self.shutdown_event = threading.Event()
        self.parser = parser or SchemaParser()

        self.system = self._setup_system(
            client,
            self.instance_name,
            system,
            name,
            description,
            version,
            icon_name,
            self.metadata,
            kwargs.get("display_name", None),
            kwargs.get("max_instances", None),
        )

        self.unique_name = "%s[%s]-%s" % (
            self.system.name,
            self.instance_name,
            self.system.version,
        )

        # Tightly manage when we're in an 'error' state, aka Brew-view is down
        self.brew_view_error_condition = threading.Condition()
        self.brew_view_down = False

        self.pool = ThreadPoolExecutor(max_workers=self.max_concurrent)
        self.admin_pool = ThreadPoolExecutor(max_workers=1)

        self.bm_client = EasyClient(
            logger=self.logger, parser=self.parser, **connection_parameters
        )

    def run(self):
        # Let Beergarden know about our system and instance
        self._initialize()

        self.logger.debug("Creating and starting admin queue consumer")
        self.admin_consumer = self._create_admin_consumer()
        self.admin_consumer.start()

        self.logger.debug("Creating and starting request queue consumer")
        self.request_consumer = self._create_standard_consumer()
        self.request_consumer.start()

        self.logger.debug("Creating and starting connection poll thread")
        self.connection_poll_thread = self._create_connection_poll_thread()
        self.connection_poll_thread.start()

        self.logger.info("Plugin %s has started", self.unique_name)

        try:
            while not self.shutdown_event.wait(0.1):
                if (
                    not self.admin_consumer.isAlive()
                    and not self.admin_consumer.shutdown_event.is_set()
                ):
                    self.logger.warning(
                        "Looks like admin consumer has died - attempting to " "restart"
                    )
                    self.admin_consumer = self._create_admin_consumer()
                    self.admin_consumer.start()

                if (
                    not self.request_consumer.isAlive()
                    and not self.request_consumer.shutdown_event.is_set()
                ):
                    self.logger.warning(
                        "Looks like request consumer has died - attempting to" "restart"
                    )
                    self.request_consumer = self._create_standard_consumer()
                    self.request_consumer.start()

                if not self.connection_poll_thread.isAlive():
                    self.logger.warning(
                        "Looks like connection poll thread has died - "
                        "attempting to restart"
                    )
                    self.connection_poll_thread = self._create_connection_poll_thread()
                    self.connection_poll_thread.start()

                if (
                    self.request_consumer.shutdown_event.is_set()
                    and self.admin_consumer.shutdown_event.is_set()
                ):
                    self.shutdown_event.set()

        except KeyboardInterrupt:
            self.logger.debug("Received KeyboardInterrupt - shutting down")
        except Exception as ex:
            self.logger.error("Event loop terminated unexpectedly - shutting down")
            self.logger.exception(ex)

        self.logger.debug("About to shut down plugin %s", self.unique_name)
        self._shutdown()

        self.logger.info("Plugin %s has terminated", self.unique_name)

    def process_message(self, target, request, headers):
        """Process a message. Intended to be run on an Executor.

        :param target: The object to invoke received commands on.
            (self or self.client)
        :param request: The parsed Request object
        :param headers: Dictionary of headers from the
            `brewtils.request_consumer.RequestConsumer`
        :return: None
        """
        request.status = "IN_PROGRESS"
        self._update_request(request, headers)

        try:
            # Set request context so this request will be the parent of any
            # generated requests and update status We also need the host/port of
            #  the current plugin. We currently don't support parent/child
            # requests across different servers.
            request_context.current_request = request
            request_context.bg_host = self.bg_host
            request_context.bg_port = self.bg_port

            output = self._invoke_command(target, request)
        except Exception as ex:
            self.logger.log(
                getattr(ex, "_bg_error_log_level", logging.ERROR),
                "Plugin %s raised an exception while processing request %s: %s",
                self.unique_name,
                str(request),
                ex,
                exc_info=not getattr(ex, "_bg_suppress_stacktrace", False),
            )
            request.status = "ERROR"
            request.output = self._format_error_output(request, ex)
            request.error_class = type(ex).__name__
        else:
            request.status = "SUCCESS"
            request.output = self._format_output(output)

        self._update_request(request, headers)

    def process_request_message(self, message, headers):
        """Processes a message from a RequestConsumer

        :param message: A valid string representation of a
            `brewtils.models.Request`
        :param headers: A dictionary of headers from the
            `brewtils.request_consumer.RequestConsumer`
        :return: A `concurrent.futures.Future`
        """

        request = self._pre_process(message)

        # This message has already been processed, all it needs to do is update
        if request.status in Request.COMPLETED_STATUSES:
            return self.pool.submit(self._update_request, request, headers)
        else:
            return self.pool.submit(self.process_message, self.client, request, headers)

    def process_admin_message(self, message, headers):

        # Admin requests won't have a system, so don't verify it
        request = self._pre_process(message, verify_system=False)

        return self.admin_pool.submit(self.process_message, self, request, headers)

    def _pre_process(self, message, verify_system=True):

        if self.shutdown_event.is_set():
            raise RequestProcessingError(
                "Unable to process message - currently shutting down"
            )

        try:
            request = self.parser.parse_request(message, from_string=True)
        except Exception as ex:
            self.logger.exception(
                "Unable to parse message body: {0}. Exception: {1}".format(message, ex)
            )
            raise DiscardMessageException("Error parsing message body")

        if (
            verify_system
            and request.command_type
            and request.command_type.upper() != "EPHEMERAL"
            and request.system.upper() != self.system.name.upper()
        ):
            raise DiscardMessageException(
                "Received message for a different system {0}".format(
                    request.system.upper()
                )
            )

        return request

    def _initialize(self):
        self.logger.debug("Initializing plugin %s", self.unique_name)

        # TODO: We should use self.bm_client.upsert_system once it is supported
        existing_system = self.bm_client.find_unique_system(
            name=self.system.name, version=self.system.version
        )

        if existing_system:
            if not existing_system.has_instance(self.instance_name):
                if len(existing_system.instances) < existing_system.max_instances:
                    existing_system.instances.append(Instance(name=self.instance_name))
                    self.bm_client.create_system(existing_system)
                else:
                    raise PluginValidationError(
                        'Unable to create plugin with instance name "%s": '
                        'System "%s[%s]" has an instance limit of %d and '
                        "existing instances %s"
                        % (
                            self.instance_name,
                            existing_system.name,
                            existing_system.version,
                            existing_system.max_instances,
                            ", ".join(existing_system.instance_names),
                        )
                    )

            # We always update in case the metadata has changed.
            self.system = self.bm_client.update_system(
                existing_system.id,
                new_commands=self.system.commands,
                metadata=self.system.metadata,
                description=self.system.description,
                display_name=self.system.display_name,
                icon_name=self.system.icon_name,
            )
        else:
            self.system = self.bm_client.create_system(self.system)

        # Sanity check to make sure an instance with this name was registered
        if self.system.has_instance(self.instance_name):
            instance_id = self.system.get_instance(self.instance_name).id
        else:
            raise PluginValidationError(
                'Unable to find registered instance with name "%s"' % self.instance_name
            )

        self.instance = self.bm_client.initialize_instance(instance_id)

        self.queue_connection_params = self.instance.queue_info.get("connection", None)
        if self.queue_connection_params and self.queue_connection_params.get(
            "ssl", None
        ):
            self.queue_connection_params["ssl"].update(
                {
                    "ca_cert": self.ca_cert,
                    "ca_verify": self.ca_verify,
                    "client_cert": self.client_cert,
                }
            )

        self.logger.debug("Plugin %s is initialized", self.unique_name)

    def _shutdown(self):
        self.shutdown_event.set()

        self.logger.debug("About to stop message consuming")
        self.request_consumer.stop_consuming()
        self.admin_consumer.stop_consuming()

        self.logger.debug("About to wake sleeping request processing threads")
        with self.brew_view_error_condition:
            self.brew_view_error_condition.notify_all()

        self.logger.debug("Shutting down request processing pool")
        self.pool.shutdown(wait=True)
        self.logger.debug("Shutting down admin processing pool")
        self.admin_pool.shutdown(wait=True)

        self.logger.debug("Attempting to stop request queue consumer")
        self.request_consumer.stop()
        self.request_consumer.join()

        self.logger.debug("Attempting to stop admin queue consumer")
        self.admin_consumer.stop()
        self.admin_consumer.join()

        self.logger.debug("Successfully shutdown plugin {0}".format(self.unique_name))

    def _create_standard_consumer(self):
        return RequestConsumer(
            thread_name="Request Consumer",
            connection_info=self.queue_connection_params,
            amqp_url=self.instance.queue_info.get("url", None),
            queue_name=self.instance.queue_info["request"]["name"],
            on_message_callback=self.process_request_message,
            panic_event=self.shutdown_event,
            max_concurrent=self.max_concurrent,
        )

    def _create_admin_consumer(self):
        return RequestConsumer(
            thread_name="Admin Consumer",
            connection_info=self.queue_connection_params,
            amqp_url=self.instance.queue_info.get("url", None),
            queue_name=self.instance.queue_info["admin"]["name"],
            on_message_callback=self.process_admin_message,
            panic_event=self.shutdown_event,
            max_concurrent=1,
            logger=logging.getLogger("brewtils.admin_consumer"),
        )

    def _create_connection_poll_thread(self):
        connection_poll_thread = threading.Thread(target=self._connection_poll)
        connection_poll_thread.daemon = True
        return connection_poll_thread

    def _invoke_command(self, target, request):
        """Invoke the function named in request.command.

        :param target: The object to search for the function implementation.
            Will be self or self.client.
        :param request: The request to process
        :raise RequestProcessingError: The specified target does not define a
            callable implementation of request.command
        :return: The output of the function invocation
        """
        if not callable(getattr(target, request.command, None)):
            raise RequestProcessingError(
                "Could not find an implementation of command '%s'" % request.command
            )

        # It's kinda weird that we need to add the object arg only if we're
        # trying to call a function on self. In both cases the function object
        # is bound... think it has something to do with our decorators
        args = [self] if target is self else []
        return getattr(target, request.command)(*args, **request.parameters)

    def _update_request(self, request, headers):
        """Sends a Request update to beer-garden

        Ephemeral requests do not get updated, so we simply skip them.

        If brew-view appears to be down, it will wait for brew-view to come back
         up before updating.

        If this is the final attempt to update, we will attempt a known, good
        request to give some information to the user. If this attempt fails
        then we simply discard the message

        :param request: The request to update
        :param headers: A dictionary of headers from
            `brewtils.request_consumer.RequestConsumer`
        :raise RepublishMessageException: The Request update failed (any reason)
        :return: None
        """

        if request.is_ephemeral:
            sys.stdout.flush()
            return

        with self.brew_view_error_condition:

            self._wait_for_brew_view_if_down(request)

            try:
                if not self._should_be_final_attempt(headers):
                    self._wait_if_not_first_attempt(headers)
                    self.bm_client.update_request(
                        request.id,
                        status=request.status,
                        output=request.output,
                        error_class=request.error_class,
                    )
                else:
                    self.bm_client.update_request(
                        request.id,
                        status="ERROR",
                        output="We tried to update the request, but it failed "
                        "too many times. Please check the plugin logs "
                        "to figure out why the request update failed. "
                        "It is possible for this request to have "
                        "succeeded, but we cannot update beer-garden "
                        "with that information.",
                        error_class="BGGivesUpError",
                    )
            except Exception as ex:
                self._handle_request_update_failure(request, headers, ex)
            finally:
                sys.stdout.flush()

    def _wait_if_not_first_attempt(self, headers):
        if headers.get("retry_attempt", 0) > 0:
            time_to_sleep = min(
                headers.get("time_to_wait", self.starting_timeout), self.max_timeout
            )
            self.shutdown_event.wait(time_to_sleep)

    def _handle_request_update_failure(self, request, headers, exc):

        # If brew-view is down, we always want to try again
        # Yes, even if it is the 'final_attempt'
        if isinstance(exc, (RequestsConnectionError, RestConnectionError)):
            self.brew_view_down = True
            self.logger.error(
                "Error updating request status: {0} exception: {1}".format(
                    request.id, exc
                )
            )
            raise RepublishRequestException(request, headers)

        elif isinstance(exc, RestClientError):
            message = (
                "Error updating request {0} and it is a client error. Probable "
                "cause is that this request is already updated. In which case, "
                "ignore this message. If request {0} did not complete, please "
                "file an issue. Discarding request to avoid an infinte loop. "
                "exception: {1}".format(request.id, exc)
            )
            self.logger.error(message)
            raise DiscardMessageException(message)

        # Time to discard the message because we've given up
        elif self._should_be_final_attempt(headers):
            message = (
                "Could not update request {0} even with a known good status, "
                "output and error_class. We have reached the final attempt and "
                "will now discard the message. Attempted to process this "
                "message {1} times".format(request.id, headers["retry_attempt"])
            )
            self.logger.error(message)
            raise DiscardMessageException(message)

        else:
            self._update_retry_attempt_information(headers)
            self.logger.exception(
                "Error updating request (Attempt #{0}: request: {1} exception: "
                "{2}".format(headers.get("retry_attempt", 0), request.id, exc)
            )
            raise RepublishRequestException(request, headers)

    def _update_retry_attempt_information(self, headers):
        headers["retry_attempt"] = headers.get("retry_attempt", 0) + 1
        headers["time_to_wait"] = min(
            headers.get("time_to_wait", self.starting_timeout // 2) * 2,
            self.max_timeout,
        )

    def _should_be_final_attempt(self, headers):
        if self.max_attempts <= 0:
            return False

        return self.max_attempts <= headers.get("retry_attempt", 0)

    def _wait_for_brew_view_if_down(self, request):
        if self.brew_view_down and not self.shutdown_event.is_set():
            self.logger.warning(
                "Currently unable to communicate with Brew-view, about to wait "
                "until connection is reestablished to update request %s",
                request.id,
            )
            self.brew_view_error_condition.wait()

    def _start(self, request):
        """Handle start message by marking this instance as running.

        :param request: The start message
        :return: Success output message
        """
        self.instance = self.bm_client.update_instance_status(
            self.instance.id, "RUNNING"
        )

        return "Successfully started plugin"

    def _stop(self, request):
        """Handle stop message by marking this instance as stopped.

        :param request: The stop message
        :return: Success output message
        """
        self.shutdown_event.set()
        self.instance = self.bm_client.update_instance_status(
            self.instance.id, "STOPPED"
        )

        return "Successfully stopped plugin"

    def _status(self, request):
        """Handle status message by sending a heartbeat.

        :param request: The status message
        :return: None
        """
        with self.brew_view_error_condition:
            if not self.brew_view_down:
                try:
                    self.bm_client.instance_heartbeat(self.instance.id)
                except (RequestsConnectionError, RestConnectionError):
                    self.brew_view_down = True
                    raise

    def _setup_max_concurrent(self, multithreaded, max_concurrent):
        """Determine correct max_concurrent value.
        Will be unnecessary when multithreaded flag is removed."""
        if multithreaded is not None:
            warnings.warn(
                "Keyword argument 'multithreaded' is deprecated and will be "
                "removed in version 3.0, please use 'max_concurrent' instead.",
                DeprecationWarning,
                stacklevel=2,
            )

            # Both multithreaded and max_concurrent kwargs explicitly set
            # check for mutually exclusive settings
            if max_concurrent is not None:
                if multithreaded is True and max_concurrent == 1:
                    self.logger.warning(
                        "Plugin created with multithreaded=True and "
                        "max_concurrent=1, ignoring 'multithreaded' argument"
                    )
                elif multithreaded is False and max_concurrent > 1:
                    self.logger.warning(
                        "Plugin created with multithreaded=False and "
                        "max_concurrent>1, ignoring 'multithreaded' argument"
                    )

                return max_concurrent
            else:
                return 5 if multithreaded else 1
        else:
            if max_concurrent is None:
                warnings.warn(
                    "Heads up - in 3.0 the default plugin behavior is changing "
                    "from handling requests one at a time to handling multiple "
                    "at once. If this plugin needs to maintain the old "
                    "behavior just set max_concurrent=1 when creating the "
                    "plugin.",
                    PendingDeprecationWarning,
                    stacklevel=2,
                )
                return 1

            return max_concurrent

    def _setup_system(
        self,
        client,
        inst_name,
        system,
        name,
        description,
        version,
        icon_name,
        metadata,
        display_name,
        max_instances,
    ):
        if system:
            if (
                name
                or description
                or version
                or icon_name
                or display_name
                or max_instances
            ):
                raise ValidationError(
                    "Sorry, you can't specify a system as well as system "
                    "creation helper keywords (name, description, version, "
                    "max_instances, display_name, and icon_name)"
                )

            if client._bg_name or client._bg_version:
                raise ValidationError(
                    "Sorry, you can't specify a system as well as system "
                    "info in the @system decorator (bg_name, bg_version)"
                )

            if not system.instances:
                raise ValidationError(
                    "Explicit system definition requires explicit instance "
                    "definition (use instances=[Instance(name='default')] for "
                    "default behavior)"
                )

            if not system.max_instances:
                system.max_instances = len(system.instances)

        else:
            name = name or os.environ.get("BG_NAME", None) or client._bg_name
            version = (
                version or os.environ.get("BG_VERSION", None) or client._bg_version
            )

            if client.__doc__ and not description:
                description = self.client.__doc__.split("\n")[0]

            system = System(
                name=name,
                description=description,
                version=version,
                icon_name=icon_name,
                commands=client._commands,
                max_instances=max_instances or 1,
                instances=[Instance(name=inst_name)],
                metadata=metadata,
                display_name=display_name,
            )

        return system

    def _connection_poll(self):
        """Periodically attempt to re-connect to beer-garden"""

        while not self.shutdown_event.wait(5):
            with self.brew_view_error_condition:
                if self.brew_view_down:
                    try:
                        self.bm_client.get_version()
                    except Exception:
                        self.logger.debug("Attempt to reconnect to Brew-view failed")
                    else:
                        self.logger.info(
                            "Brew-view connection reestablished, about to "
                            "notify any waiting requests"
                        )
                        self.brew_view_down = False
                        self.brew_view_error_condition.notify_all()

    @staticmethod
    def _format_error_output(request, exc):
        if request.is_json:
            return parse_exception_as_json(exc)
        else:
            return str(exc)

    @staticmethod
    def _format_output(output):
        """Formats output from Plugins to prevent validation errors"""

        if isinstance(output, six.string_types):
            return output

        try:
            return json.dumps(output)
        except (TypeError, ValueError):
            return str(output)


# Alias old name
PluginBase = Plugin


class RemotePlugin(Plugin):
    pass



# -*- coding: utf-8 -*-

from lark import Lark, Transformer

# Lark added some new errors in later versions
# Lark also moved around their error in 0.6.0
try:
    from lark import ParseError
except ImportError:
    from lark.common import ParseError

try:
    from lark import GrammarError, LexError
except ImportError:
    GrammarError = ParseError
    LexError = ParseError


choices_grammar = r"""
    func: CNAME [func_args]
    url: ADDRESS [url_args]
    reference: ref

    func_args: "(" [arg_pair ("," arg_pair)*] ")"
    url_args: "?" arg_pair ("&" arg_pair)*

    arg_pair: CNAME "=" ref
    ?ref: "${" CNAME "}"

    ADDRESS: /^http[^\?]*/

    %import common.CNAME
    %import common.WS
    %ignore WS
"""

parsers = {
    "func": Lark(choices_grammar, start="func"),
    "url": Lark(choices_grammar, start="url"),
    "reference": Lark(choices_grammar, start="reference"),
}


class FunctionTransformer(Transformer):
    @staticmethod
    def func(s):
        return {"name": str(s[0]), "args": s[1] if len(s) > 1 else []}

    @staticmethod
    def url(s):
        return {"address": str(s[0]), "args": s[1] if len(s) > 1 else []}

    @staticmethod
    def reference(s):
        return str(s[0])

    @staticmethod
    def arg_pair(s):
        return str(s[0]), str(s[1])

    func_args = list
    url_args = list


def parse(input_string, parse_as=None):
    """Attempt to parse a string into a choices dictionary.

    :param input_string: The string to parse
    :param parse_as: String specifying how to parse `input_string`. Valid values are 'func' or
        'url'. Will try all valid values if None.
    :return: A dictionary containing the results of the parse
    :raise lark.common.ParseError: The parser was not able to find a valid parsing of `input_string`
    """

    def _parse(_input_string, _parser):
        try:
            return FunctionTransformer().transform(_parser.parse(_input_string))
        except (GrammarError, LexError, ParseError) as e:
            raise ParseError(e)

    if parse_as is not None:
        return _parse(input_string, parsers[parse_as])
    else:
        for parser in parsers.values():
            try:
                return _parse(input_string, parser)
            except ParseError:
                continue

    raise ParseError('Unable to successfully parse input "%s"' % input_string)



# -*- coding: utf-8 -*-

import functools
import inspect
import json
import os
import sys
import types
from io import open

import requests
import six
import wrapt

try:
    from lark import ParseError
except ImportError:
    from lark.common import ParseError

from brewtils.choices import parse
from brewtils.errors import PluginParamError
from brewtils.models import Command, Parameter, Choices

if sys.version_info.major == 2:
    from funcsigs import signature, Parameter as InspectParameter
else:
    from inspect import signature, Parameter as InspectParameter

__all__ = [
    "system",
    "parameter",
    "parameters",
    "command",
    "command_registrar",
    "plugin_param",
    "register",
]


# The wrapt module has a cool feature where you can disable wrapping a decorated function,
# instead just using the original function. This is pretty much exactly what we want - we
# aren't using decorators for their 'real' purpose of wrapping a function, we just want to add
# some metadata to the function object. So we'll disable the wrapping normally, but we need to
# test that enabling the wrapping would work.
_wrap_functions = False


def system(cls=None, bg_name=None, bg_version=None):
    """Class decorator that marks a class as a beer-garden System

    Creates some properties on the class:
      * ``_commands``: holds all registered commands
      * ``_name``: an optional system name
      * ``_version``: an optional system version
      * ``_current_request``: Reference to the currently executing request

    Args:
        cls: The class to decorated
        bg_name: Optional plugin name
        bg_version: Optional plugin version

    Returns:
        The decorated class

    """
    if cls is None:
        return functools.partial(system, bg_name=bg_name, bg_version=bg_version)

    import brewtils.plugin

    commands = []
    for method_name in dir(cls):
        method = getattr(cls, method_name)
        method_command = getattr(method, "_command", None)
        if method_command:
            commands.append(method_command)

    cls._commands = commands
    cls._bg_name = bg_name
    cls._bg_version = bg_version
    cls._current_request = property(
        lambda self: brewtils.plugin.request_context.current_request
    )

    return cls


def command(
    _wrapped=None,
    command_type="ACTION",
    output_type="STRING",
    schema=None,
    form=None,
    template=None,
    icon_name=None,
    description=None,
):
    """Decorator that marks a function as a beer-garden command

    For example:

    .. code-block:: python

        @command(output_type='JSON')
        def echo_json(self, message):
            return message

    :param _wrapped: The function to decorate. This is handled as a positional argument and
        shouldn't be explicitly set.
    :param command_type: The command type. Valid options are Command.COMMAND_TYPES.
    :param output_type: The output type. Valid options are Command.OUTPUT_TYPES.
    :param schema: A custom schema definition.
    :param form: A custom form definition.
    :param template: A custom template definition.
    :param icon_name: The icon name. Should be either a FontAwesome or a Glyphicon name.
    :param description: The command description. Will override the function's docstring.
    :return: The decorated function.
    """
    if _wrapped is None:
        return functools.partial(
            command,
            command_type=command_type,
            output_type=output_type,
            schema=schema,
            form=form,
            template=template,
            icon_name=icon_name,
            description=description,
        )

    generated_command = _generate_command_from_function(_wrapped)
    generated_command.command_type = command_type
    generated_command.output_type = output_type
    generated_command.icon_name = icon_name

    if description:
        generated_command.description = description

    resolved_mod = _resolve_display_modifiers(
        _wrapped, generated_command.name, schema=schema, form=form, template=template
    )
    generated_command.schema = resolved_mod["schema"]
    generated_command.form = resolved_mod["form"]
    generated_command.template = resolved_mod["template"]

    func_command = getattr(_wrapped, "_command", None)
    if func_command:
        _update_func_command(func_command, generated_command)
    else:
        _wrapped._command = generated_command

    @wrapt.decorator(enabled=_wrap_functions)
    def wrapper(_double_wrapped, _, _args, _kwargs):
        return _double_wrapped(*_args, **_kwargs)

    return wrapper(_wrapped)


def parameter(
    _wrapped=None,
    key=None,
    type=None,
    multi=None,
    display_name=None,
    optional=None,
    default=None,
    description=None,
    choices=None,
    nullable=None,
    maximum=None,
    minimum=None,
    regex=None,
    is_kwarg=None,
    model=None,
    form_input_type=None,
):
    """Decorator that enables Parameter specifications for a beer-garden Command

    This decorator is intended to be used when more specification is desired for a Parameter.

    For example::

        @parameter(key="message", description="Message to echo", optional=True, type="String",
                   default="Hello, World!")
        def echo(self, message):
            return message

    :param _wrapped: The function to decorate. This is handled as a positional argument and
        shouldn't be explicitly set.
    :param key: String specifying the parameter identifier. Must match an argument name of the
        decorated function.
    :param type: String indicating the type to use for this parameter.
    :param multi: Boolean indicating if this parameter is a multi. See documentation for
        discussion of what this means.
    :param display_name: String that will be displayed as a label in the user interface.
    :param optional: Boolean indicating if this parameter must be specified.
    :param default: The value this parameter will be assigned if not overridden when creating a
        request.
    :param description: An additional string that will be displayed in the user interface.
    :param choices: List or dictionary specifying allowed values. See documentation for more
        information.
    :param nullable: Boolean indicating if this parameter is allowed to be null.
    :param maximum: Integer indicating the maximum value of the parameter.
    :param minimum: Integer indicating the minimum value of the parameter.
    :param regex: String describing a regular expression constraint on the parameter.
    :param is_kwarg: Boolean indicating if this parameter is meant to be part of the decorated
        function's kwargs.
    :param model: Class to be used as a model for this parameter. Must be a Python type object,
        not an instance.
    :param form_input_type: Only used for string fields. Changes the form input field
        (e.g. textarea)
    :return: The decorated function.
    """
    if _wrapped is None:
        return functools.partial(
            parameter,
            key=key,
            type=type,
            multi=multi,
            display_name=display_name,
            optional=optional,
            default=default,
            description=description,
            choices=choices,
            nullable=nullable,
            maximum=maximum,
            minimum=minimum,
            regex=regex,
            is_kwarg=is_kwarg,
            model=model,
            form_input_type=form_input_type,
        )

    # Create a command object if one isn't already associated
    cmd = getattr(_wrapped, "_command", None)
    if not cmd:
        cmd = _generate_command_from_function(_wrapped)
        _wrapped._command = cmd

    # Every parameter needs a key, so stop that right here
    if key is None:
        raise PluginParamError(
            "Found a parameter definition without a key for " "command '%s'" % cmd.name
        )

    # If the command doesn't already have a parameter with this key then the
    # method doesn't have an explicit keyword argument with <key> as the name.
    # That's only OK if this parameter is meant to be part of the **kwargs.
    param = cmd.get_parameter_by_key(key)
    if param is None:
        if is_kwarg:
            param = Parameter(key=key, optional=False)
            cmd.parameters.append(param)
        else:
            raise PluginParamError(
                "Parameter '%s' was not an explicit keyword argument for "
                "command '%s' and was not marked as part of kwargs (should "
                "is_kwarg be True?)" % (key, cmd.name)
            )

    # Next, fail if the param is_kwarg=True and the method doesn't have a **kwargs
    if is_kwarg:
        kwarg_declared = False
        for p in signature(_wrapped).parameters.values():
            if p.kind == InspectParameter.VAR_KEYWORD:
                kwarg_declared = True
                break

        if not kwarg_declared:
            raise PluginParamError(
                "Parameter '%s' of command '%s' was declared as a kwarg argument "
                "(is_kwarg=True) but the command method does not declare a **kwargs "
                "argument" % (key, cmd.name)
            )

    # Update parameter definition with the plugin_param arguments
    param.type = _format_type(param.type if type is None else type)
    param.multi = param.multi if multi is None else multi
    param.display_name = param.display_name if display_name is None else display_name
    param.optional = param.optional if optional is None else optional
    param.default = param.default if default is None else default
    param.description = param.description if description is None else description
    param.choices = param.choices if choices is None else choices
    param.nullable = param.nullable if nullable is None else nullable
    param.maximum = param.maximum if maximum is None else maximum
    param.minimum = param.minimum if minimum is None else minimum
    param.regex = param.regex if regex is None else regex
    param.form_input_type = (
        param.form_input_type if form_input_type is None else form_input_type
    )

    param.choices = _format_choices(param.choices)

    # Model is another special case - it requires its own handling
    if model is not None:
        param.type = "Dictionary"
        param.parameters = _generate_nested_params(model)

        # If the model is not nullable and does not have a default we will try
        # to generate a one using the defaults defined on the model parameters
        if not param.nullable and not param.default:
            param.default = {}
            for nested_param in param.parameters:
                if nested_param.default:
                    param.default[nested_param.key] = nested_param.default

    @wrapt.decorator(enabled=_wrap_functions)
    def wrapper(_double_wrapped, _, _args, _kwargs):
        return _double_wrapped(*_args, **_kwargs)

    return wrapper(_wrapped)


def parameters(*args):
    """Specify multiple Parameter definitions at once

    This can be useful for commands which have a large number of complicated
    parameters but aren't good candidates for a Model.

    .. code-block:: python

        @parameter(**params[cmd1][param1])
        @parameter(**params[cmd1][param2])
        @parameter(**params[cmd1][param3])
        def cmd1(self, **kwargs):
            pass

    Can become:

    .. code-block:: python

        @parameters(params[cmd1])
        def cmd1(self, **kwargs):
            pass

    Args:
        *args (iterable): Positional arguments
            The first (and only) positional argument must be a list containing
            dictionaries that describe parameters.

    Returns:
        func: The decorated function
    """
    if len(args) == 1:
        return functools.partial(parameters, args[0])
    elif len(args) != 2:
        raise PluginParamError("@parameters takes a single argument")

    if not isinstance(args[1], types.FunctionType):
        raise PluginParamError("@parameters must be applied to a function")

    try:
        for param in args[0]:
            parameter(args[1], **param)
    except TypeError:
        raise PluginParamError("@parameters arg must be an iterable of dictionaries")

    @wrapt.decorator(enabled=_wrap_functions)
    def wrapper(_double_wrapped, _, _args, _kwargs):
        return _double_wrapped(*_args, **_kwargs)

    return wrapper(args[1])


def _update_func_command(func_command, generated_command):
    """Updates the current function's command with info, (will not override plugin_params)"""
    func_command.name = generated_command.name
    func_command.description = generated_command.description
    func_command.command_type = generated_command.command_type
    func_command.output_type = generated_command.output_type
    func_command.schema = generated_command.schema
    func_command.form = generated_command.form
    func_command.template = generated_command.template
    func_command.icon_name = generated_command.icon_name


def _generate_command_from_function(func):
    """Generates a Command from a function. Uses first line of pydoc as the description."""
    # Required for Python 2/3 compatibility
    if hasattr(func, "func_name"):
        command_name = func.func_name
    else:
        command_name = func.__name__

    # Required for Python 2/3 compatibility
    if hasattr(func, "func_doc"):
        docstring = func.func_doc
    else:
        docstring = func.__doc__

    return Command(
        name=command_name,
        description=docstring.split("\n")[0] if docstring else None,
        parameters=_generate_params_from_function(func),
    )


def _generate_params_from_function(func):
    """Generate Parameters from function arguments.
    Will set the Parameter key, default value, and optional value."""
    parameters_to_return = []

    code = six.get_function_code(func)
    function_arguments = list(code.co_varnames or [])[: code.co_argcount]
    function_defaults = list(six.get_function_defaults(func) or [])

    while len(function_defaults) != len(function_arguments):
        function_defaults.insert(0, None)

    for index, param_name in enumerate(function_arguments):
        # Skip Self or Class reference
        if index == 0 and isinstance(func, types.FunctionType):
            continue

        default = function_defaults[index]
        optional = False if default is None else True

        parameters_to_return.append(
            Parameter(key=param_name, default=default, optional=optional)
        )

    return parameters_to_return


def _generate_nested_params(model_class):
    """Generates Nested Parameters from a Model Class"""
    parameters_to_return = []
    for parameter_definition in model_class.parameters:
        key = parameter_definition.key
        parameter_type = parameter_definition.type
        multi = parameter_definition.multi
        display_name = parameter_definition.display_name
        optional = parameter_definition.optional
        default = parameter_definition.default
        description = parameter_definition.description
        nullable = parameter_definition.nullable
        maximum = parameter_definition.maximum
        minimum = parameter_definition.minimum
        regex = parameter_definition.regex

        choices = _format_choices(parameter_definition.choices)

        nested_parameters = []
        if parameter_definition.parameters:
            parameter_type = "Dictionary"
            for nested_class in parameter_definition.parameters:
                nested_parameters = _generate_nested_params(nested_class)

        parameters_to_return.append(
            Parameter(
                key=key,
                type=parameter_type,
                multi=multi,
                display_name=display_name,
                optional=optional,
                default=default,
                description=description,
                choices=choices,
                parameters=nested_parameters,
                nullable=nullable,
                maximum=maximum,
                minimum=minimum,
                regex=regex,
            )
        )
    return parameters_to_return


def _resolve_display_modifiers(
    wrapped, command_name, schema=None, form=None, template=None
):
    def _load_from_url(url):
        response = requests.get(url)
        if response.headers.get("content-type", "").lower() == "application/json":
            return json.loads(response.text)
        return response.text

    def _load_from_path(path):
        current_dir = os.path.dirname(inspect.getfile(wrapped))
        file_path = os.path.abspath(os.path.join(current_dir, path))

        with open(file_path, "r") as definition_file:
            return definition_file.read()

    resolved = {}

    for key, value in {"schema": schema, "form": form, "template": template}.items():

        if isinstance(value, six.string_types):
            try:
                if value.startswith("http"):
                    resolved[key] = _load_from_url(value)

                elif value.startswith("/") or value.startswith("."):
                    loaded_value = _load_from_path(value)
                    resolved[key] = (
                        loaded_value if key == "template" else json.loads(loaded_value)
                    )

                elif key == "template":
                    resolved[key] = value

                else:
                    raise PluginParamError(
                        "%s specified for command '%s' was not a "
                        "definition, file path, or URL" % (key, command_name)
                    )
            except Exception as ex:
                raise PluginParamError(
                    "Error reading %s definition from '%s' for command "
                    "'%s': %s" % (key, value, command_name, ex)
                )

        elif value is None or (key in ["schema", "form"] and isinstance(value, dict)):
            resolved[key] = value

        elif key == "form" and isinstance(value, list):
            resolved[key] = {"type": "fieldset", "items": value}

        else:
            raise PluginParamError(
                "%s specified for command '%s' was not a definition, "
                "file path, or URL" % (key, command_name)
            )

    return resolved


def _format_type(param_type):
    if param_type == str:
        return "String"
    elif param_type == int:
        return "Integer"
    elif param_type == float:
        return "Float"
    elif param_type == bool:
        return "Boolean"
    elif param_type == dict:
        return "Dictionary"
    else:
        return param_type


def _format_choices(choices):
    def determine_display(display_value):
        if isinstance(display_value, six.string_types):
            return "typeahead"

        return "select" if len(display_value) <= 50 else "typeahead"

    def determine_type(type_value):
        if isinstance(type_value, (list, dict)):
            return "static"
        elif type_value.startswith("http"):
            return "url"
        else:
            return "command"

    if not choices:
        return None

    if not isinstance(choices, (list, six.string_types, dict)):
        raise PluginParamError(
            "Invalid 'choices' provided. Must be a list, dictionary or string."
        )

    elif isinstance(choices, dict):
        if not choices.get("value"):
            raise PluginParamError(
                "No 'value' provided for choices. You must at least "
                "provide valid values."
            )

        value = choices.get("value")
        display = choices.get("display", determine_display(value))
        choice_type = choices.get("type")
        strict = choices.get("strict", True)

        if choice_type is None:
            choice_type = determine_type(value)
        elif choice_type not in Choices.TYPES:
            raise PluginParamError(
                "Invalid choices type '%s' - Valid type options are %s"
                % (choice_type, Choices.TYPES)
            )
        else:
            if (
                (
                    choice_type == "command"
                    and not isinstance(value, (six.string_types, dict))
                )
                or (choice_type == "url" and not isinstance(value, six.string_types))
                or (choice_type == "static" and not isinstance(value, (list, dict)))
            ):
                allowed_types = {
                    "command": "('string', 'dictionary')",
                    "url": "('string')",
                    "static": "('list', 'dictionary)",
                }
                raise PluginParamError(
                    "Invalid choices value type '%s' - Valid value types for "
                    "choice type '%s' are %s"
                    % (type(value), choice_type, allowed_types[choice_type])
                )

        if display not in Choices.DISPLAYS:
            raise PluginParamError(
                "Invalid choices display '%s' - Valid display options are %s"
                % (display, Choices.DISPLAYS)
            )
    else:
        value = choices
        display = determine_display(value)
        choice_type = determine_type(value)
        strict = True

    # Now parse out type-specific aspects
    unparsed_value = ""
    try:
        if choice_type == "command":
            if isinstance(value, six.string_types):
                unparsed_value = value
            else:
                unparsed_value = value["command"]

            details = parse(unparsed_value, parse_as="func")
        elif choice_type == "url":
            unparsed_value = value
            details = parse(unparsed_value, parse_as="url")
        else:
            if isinstance(value, dict):
                unparsed_value = choices.get("key_reference")
                if unparsed_value is None:
                    raise PluginParamError(
                        "Specifying a static choices dictionary requires a "
                        '"key_reference" field with a reference to another '
                        'parameter ("key_reference": "${param_key}")'
                    )

                details = {"key_reference": parse(unparsed_value, parse_as="reference")}
            else:
                details = {}
    except ParseError:
        raise PluginParamError(
            "Invalid choices definition - Unable to parse '%s'" % unparsed_value
        )

    return Choices(
        type=choice_type, display=display, value=value, strict=strict, details=details
    )


# Alias the old names for compatibility
command_registrar = system
plugin_param = parameter
register = command



# -*- coding: utf-8 -*-


def normalize_url_prefix(url_prefix):
    """Enforce a consistent URL representation

    The normalized prefix will begin and end with '/'. If there is no prefix
    the normalized form will be '/'.

    Examples:
        INPUT       NORMALIZED
        None            '/'
        ''              '/'
        '/'             '/'
        'example'       '/example/'
        '/example'      '/example/'
        'example/'      '/example/'
        '/example/'     '/example/'

    Args:
        url_prefix (str): The prefix

    Returns:
        str: The normalized prefix
    """
    if url_prefix in (None, "/", ""):
        return "/"

    new_url_prefix = ""

    # Make string begin with /
    if not url_prefix.startswith("/"):
        new_url_prefix += "/"

    new_url_prefix += url_prefix

    # Make string end with /
    if not url_prefix.endswith("/"):
        new_url_prefix += "/"

    return new_url_prefix



# -*- coding: utf-8 -*-
"""Brewtils Logging Utilities

This module streamlines loading logging configuration from Beergarden.

Example:
    To use this just call ``configure_logging`` sometime before you initialize
    your Plugin object:

    .. code-block:: python

        from brewtils import configure_logging, get_connection_info, Plugin

        # Load BG connection info from environment and command line args
        connection_info = get_connection_info(sys.argv[1:])

        configure_logging(system_name='systemX', **connection_info)

        plugin = Plugin(
            my_client,
            name='systemX,
            version='0.0.1',
            **connection_info
        )
        plugin.run()
"""

import copy
import logging.config
import os
import warnings

import brewtils

# Loggers to always use. These are things that generally,
# people do not want to see and/or are too verbose.
DEFAULT_LOGGERS = {
    "pika": {"level": "ERROR"},
    "requests.packages.urllib3.connectionpool": {"level": "WARN"},
    "yapconf": {"level": "WARN"},
}

# A simple default format/formatter. Generally speaking, the API should return
# formatters, but since users can configure their logging it's better if the
# formatter has a logical backup.
DEFAULT_FORMAT = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
DEFAULT_FORMATTERS = {"default": {"format": DEFAULT_FORMAT}}

# A simple default handler. Generally speaking, the API should return
# handlers, but since users can configure their logging it's better if the
# handler has a logical backup.
DEFAULT_HANDLERS = {
    "default": {
        "class": "logging.StreamHandler",
        "formatter": "default",
        "stream": "ext://sys.stdout",
    }
}

# The template that plugins will use to log
DEFAULT_PLUGIN_LOGGING_TEMPLATE = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {},
    "handlers": {},
    "loggers": DEFAULT_LOGGERS,
}

# If no logging was configured, this will be used as the logging configuration
DEFAULT_LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": DEFAULT_FORMATTERS,
    "handlers": DEFAULT_HANDLERS,
    "loggers": DEFAULT_LOGGERS,
    "root": {"level": os.environ.get("BG_LOG_LEVEL", "INFO"), "handlers": ["default"]},
}


def configure_logging(system_name=None, **kwargs):
    """Load and enable a logging configuration from Beergarden

    NOTE: This method will overwrite the current logging configuration.

    Args:
        system_name: Name of the system to load
        **kwargs: Beergarden connection parameters

    Returns:
        None
    """
    config = get_logging_config(system_name=system_name, **kwargs)
    logging.config.dictConfig(config)


def get_logging_config(system_name=None, **kwargs):
    """Retrieve a logging configuration from Beergarden

    Args:
        system_name: Name of the system to load
        **kwargs: Beergarden connection parameters

    Returns:
        dict: The logging configuration for the specified system
    """
    config = brewtils.get_easy_client(**kwargs).get_logging_config(system_name)

    return convert_logging_config(config)


def convert_logging_config(logging_config):
    """Transform a LoggingConfig object into a Python logging configuration

    Args:
        logging_config: Beergarden logging config

    Returns:
        dict: The logging configuration
    """
    config_to_return = copy.deepcopy(DEFAULT_PLUGIN_LOGGING_TEMPLATE)

    if logging_config.handlers:
        handlers = logging_config.handlers
    else:
        handlers = copy.deepcopy(DEFAULT_HANDLERS)
    config_to_return["handlers"] = handlers

    if logging_config.formatters:
        formatters = logging_config.formatters
    else:
        formatters = copy.deepcopy(DEFAULT_FORMATTERS)
    config_to_return["formatters"] = formatters

    config_to_return["root"] = {
        "level": logging_config.level,
        "handlers": list(config_to_return["handlers"]),
    }

    return config_to_return


def setup_logger(
    bg_host, bg_port, system_name, ca_cert=None, client_cert=None, ssl_enabled=None
):
    """DEPRECATED: Set Python logging to use configuration from Beergarden API

    This method is deprecated - consider using :func:`configure_logging`

    This method will overwrite the current logging configuration.

    Args:
        bg_host (str): Beergarden host
        bg_port (int): Beergarden port
        system_name (str): Name of the system
        ca_cert (str): Path to CA certificate file
        client_cert (str): Path to client certificate file
        ssl_enabled (bool): Use SSL when connection to Beergarden

    Returns: None
    """
    warnings.warn(
        "This function is deprecated and will be removed in version "
        "4.0, please consider using 'configure_logging' instead.",
        DeprecationWarning,
        stacklevel=2,
    )

    config = get_python_logging_config(
        bg_host=bg_host,
        bg_port=bg_port,
        system_name=system_name,
        ca_cert=ca_cert,
        client_cert=client_cert,
        ssl_enabled=ssl_enabled,
    )
    logging.config.dictConfig(config)


def get_python_logging_config(
    bg_host, bg_port, system_name, ca_cert=None, client_cert=None, ssl_enabled=None
):
    """DEPRECATED: Get Beergarden's logging configuration

    This method is deprecated - consider using :func:`get_logging_config`

    Args:
        bg_host (str): Beergarden host
        bg_port (int): Beergarden port
        system_name (str): Name of the system
        ca_cert (str): Path to CA certificate file
        client_cert (str): Path to client certificate file
        ssl_enabled (bool): Use SSL when connection to Beergarden

    Returns:
        dict: The logging configuration for the specified system
    """
    warnings.warn(
        "This function is deprecated and will be removed in version "
        "4.0, please consider using 'get_logging_config' instead.",
        DeprecationWarning,
        stacklevel=2,
    )

    client = brewtils.get_easy_client(
        host=bg_host,
        port=bg_port,
        ssl_enabled=ssl_enabled,
        ca_cert=ca_cert,
        client_cert=client_cert,
    )

    logging_config = client.get_logging_config(system_name=system_name)

    return convert_logging_config(logging_config)



# -*- coding: utf-8 -*-

import logging
import warnings
from concurrent.futures import ThreadPoolExecutor
from multiprocessing import cpu_count

import time
from functools import partial
from packaging.version import parse

from brewtils.errors import (
    FetchError,
    TimeoutExceededError,
    RequestFailedError,
    ValidationError,
)
from brewtils.models import Request
from brewtils.plugin import request_context
from brewtils.rest.easy_client import EasyClient


class SystemClient(object):
    """High-level client for generating requests for a beer-garden System.

    SystemClient creation:
        This class is intended to be the main way to create beer-garden requests. Create an
        instance with beer-garden connection information (optionally including a url_prefix) and
        a system name::

            client = SystemClient(host, port, 'example_system', ssl_enabled=True, url_prefix=None)

        Pass additional keyword arguments for more granularity:

            version_constraint:
                Allows specifying a particular system version. Can be a version literal ('1.0.0')
                or the special value 'latest.' Using 'latest' will allow the the SystemClient to
                retry a request if it fails due to a missing system (see Creating Requests).

            default_instance:
                The instance name to use when creating a request if no other instance name is
                specified. Since each request must be addressed to a specific instance this is a
                convenience to prevent needing to specify the 'default' instance for each request.

            always_update:
                Always attempt to reload the system definition before making a request. This is
                useful to ensure Requests are always made against the latest version of the system.
                If not set the System definition will be loaded once (upon making the first
                request) and then only reloaded if a Request fails.

    Loading the System:
        The System definition is lazily loaded, so nothing happens until the first attempt to send
        a Request. At that point the SystemClient will query beer-garden to get a system definition
        that matches the system_name and version_constraint. If no matching system can be found a
        FetchError will be raised. If always_update was set to True this will happen
        before making each request, not just the first.

    Making a Request:
        The standard way to create and send requests is by calling object attributes::

            request = client.example_command(param_1='example_param')

        In the normal case this will block until the request completes. Request completion is
        determined by periodically polling beer-garden to check the Request status. The time
        between polling requests starts at 0.5s and doubles each time the request has still not
        completed, up to max_delay. If a timeout was specified and the Request has not completed
        within that time a ``ConnectionTimeoutError`` will be raised.

        It is also possible to create the SystemClient in non-blocking mode by specifying
        blocking=False. In this case the request creation will immediately return a Future and
        will spawn a separate thread to poll for Request completion. The max_concurrent parameter
        is used to control the maximum threads available for polling.

        .. code-block:: python

            # Create a SystemClient with blocking=False
            client = SystemClient(host, port, 'example_system', ssl_enabled=True, blocking=False)

            # Create and send 5 requests without waiting for request completion
            futures = [client.example_command(param_1=number) for number in range(5)]

            # Now wait on all requests to complete
            concurrent.futures.wait(futures)

        If the request creation process fails (e.g. the command failed validation) and
        version_constraint is 'latest' then the SystemClient will check to see if a different
        version is available, and if so it will attempt to make the request on that version.
        This is so users of the SystemClient that don't necessarily care about the target system
        version don't need to be restarted if the target system is updated.

    Tweaking beer-garden Request Parameters:
        There are several parameters that control how beer-garden routes / processes a request. To
        denote these as intended for beer-garden itself (rather than a parameter to be passed to
        the Plugin) prepend a leading underscore to the argument name.

        Sending to another instance::

            request = client.example_command(_instance_name='instance_2', param_1='example_param')

        Request with a comment::

            request = client.example_command(_comment='I'm a beer-garden comment!',
                                             param_1='example_param')

        Without the leading underscore the arguments would be treated the same as param_1 -
        another parameter to be passed to the plugin.

    :param host: beer-garden REST API hostname.
    :param port: beer-garden REST API port.
    :param system_name: The name of the system to use.
    :param version_constraint: The system version to use. Can be specific or 'latest'.
    :param default_instance: The instance to use if not specified when creating a request.
    :param always_update: Should check for a newer System version before each request.
    :param timeout: Length of time to wait for a request to complete. 'None' means wait forever.
    :param max_delay: Maximum time to wait between checking the status of a created request.
    :param api_version: beer-garden API version.
    :param ssl_enabled: Flag indicating whether to use HTTPS when communicating with beer-garden.
    :param ca_cert: beer-garden REST API server CA certificate.
    :param blocking: Block after request creation until the request completes.
    :param max_concurrent: Maximum number of concurrent requests allowed.
    :param client_cert: The client certificate to use when making requests.
    :param url_prefix: beer-garden REST API URL Prefix.
    :param ca_verify: Flag indicating whether to verify server certificate when making a request.
    :param raise_on_error: Raises an error if the request ends in an error state.
    :param username: Username for Beergarden authentication
    :param password: Password for Beergarden authentication
    :param access_token: Access token for Beergarden authentication
    :param refresh_token: Refresh token for Beergarden authentication
    :param client_timeout: Max time to will wait for server response
    """

    def __init__(
        self,
        bg_host=None,
        bg_port=None,
        system_name=None,
        version_constraint="latest",
        default_instance="default",
        always_update=False,
        timeout=None,
        max_delay=30,
        api_version=None,
        ssl_enabled=False,
        ca_cert=None,
        blocking=True,
        max_concurrent=None,
        client_cert=None,
        url_prefix=None,
        ca_verify=True,
        raise_on_error=False,
        **kwargs
    ):
        self._system_name = system_name
        self._version_constraint = version_constraint
        self._default_instance = default_instance
        self._always_update = always_update
        self._timeout = timeout
        self._max_delay = max_delay
        self._blocking = blocking
        self._raise_on_error = raise_on_error
        self._bg_host = bg_host or kwargs.get("host")
        self._bg_port = bg_port or kwargs.get("port")
        self.logger = logging.getLogger(__name__)

        self._loaded = False
        self._system = None
        self._commands = None

        # This is for Python 3.4 compatibility - max_workers MUST be non-None
        # in that version. This logic is what was added in Python 3.5
        if max_concurrent is None:
            max_concurrent = (cpu_count() or 1) * 5
        self._thread_pool = ThreadPoolExecutor(max_workers=max_concurrent)

        self._easy_client = EasyClient(
            bg_host=self._bg_host,
            bg_port=self._bg_port,
            ssl_enabled=ssl_enabled,
            api_version=api_version,
            ca_cert=ca_cert,
            client_cert=client_cert,
            url_prefix=url_prefix,
            ca_verify=ca_verify,
            **kwargs
        )

    def __getattr__(self, item):
        """Standard way to create and send beer-garden requests"""
        return self.create_bg_request(item)

    def create_bg_request(self, command_name, **kwargs):
        """Create a callable that will execute a beer-garden request when called.

        Normally you interact with the SystemClient by accessing attributes, but there could be
        certain cases where you want to create a request without sending it.

        Example::

            client = SystemClient(host, port, 'system', blocking=False)
            requests = []

            # No arguments
            requests.append(client.create_bg_request('command_1'))

            # arg_1 will be passed as a parameter
            requests.append(client.create_bg_request('command_2', arg_1='Hi!'))

            futures = [request() for request in requests]   # Calling creates and sends the request
            concurrent.futures.wait(futures)                # Wait for all the futures to complete

        :param command_name: The name of the command that will be sent.
        :param kwargs: Additional arguments to pass to send_bg_request.
        :raise AttributeError: The system does not have a command with the given command_name.
        :return: A partial that will create and execute a beer-garden request when called.
        """

        if not self._loaded or self._always_update:
            self.load_bg_system()

        if command_name in self._commands:
            return partial(
                self.send_bg_request,
                _command=command_name,
                _system_name=self._system.name,
                _system_version=self._system.version,
                _system_display=self._system.display_name,
                _output_type=self._commands[command_name].output_type,
                _instance_name=self._default_instance,
                **kwargs
            )
        else:
            raise AttributeError(
                "System '%s' version '%s' has no command named '%s'"
                % (self._system.name, self._system.version, command_name)
            )

    def send_bg_request(self, **kwargs):
        """Actually create a Request and send it to beer-garden

        .. note::
            This method is intended for advanced use only, mainly cases where you're using the
            SystemClient without a predefined System. It assumes that everything needed to
            construct the request is being passed in kwargs. If this doesn't sound like what you
            want you should check out create_bg_request.

        :param kwargs: All necessary request parameters, including beer-garden internal parameters
        :raise ValidationError: If the Request creation failed validation on the server
        :return: If the SystemClient was created with blocking=True a completed request object,
            otherwise a Future that will return the Request when it completes.
        """
        # Need to pop here, otherwise we'll try to send as a request parameter
        raise_on_error = kwargs.pop("_raise_on_error", self._raise_on_error)
        blocking = kwargs.pop("_blocking", self._blocking)
        timeout = kwargs.pop("_timeout", self._timeout)

        # If the request fails validation and the version constraint allows,
        # check for a new version and retry
        try:
            request = self._construct_bg_request(**kwargs)
            request = self._easy_client.create_request(
                request, blocking=blocking, timeout=timeout
            )
        except ValidationError:
            if self._system and self._version_constraint == "latest":
                old_version = self._system.version

                self.load_bg_system()

                if old_version != self._system.version:
                    kwargs["_system_version"] = self._system.version
                    return self.send_bg_request(**kwargs)
            raise

        # If not blocking just return the future
        if not blocking:
            return self._thread_pool.submit(
                self._wait_for_request, request, raise_on_error, timeout
            )

        # Brew-view before 2.4 doesn't support the blocking flag, so make sure
        # the request is actually complete before returning
        return self._wait_for_request(request, raise_on_error, timeout)

    def load_bg_system(self):
        """Query beer-garden for a System definition

        This method will make the query to beer-garden for a System matching the name and version
        constraints specified during SystemClient instance creation.

        If this method completes successfully the SystemClient will be ready to create and send
        Requests.

        :raise FetchError: If unable to find a matching System
        :return: None
        """

        if self._version_constraint == "latest":
            self._system = self._determine_latest(
                self._easy_client.find_systems(name=self._system_name)
            )
        else:
            self._system = self._easy_client.find_unique_system(
                name=self._system_name, version=self._version_constraint
            )

        if self._system is None:
            raise FetchError(
                "Beer-garden has no system named '%s' with a version matching '%s'"
                % (self._system_name, self._version_constraint)
            )

        self._commands = {command.name: command for command in self._system.commands}
        self._loaded = True

    def _wait_for_request(self, request, raise_on_error, timeout):
        """Poll the server until the request is completed or errors"""

        delay_time = 0.5
        total_wait_time = 0
        while request.status not in Request.COMPLETED_STATUSES:

            if timeout and total_wait_time > timeout:
                raise TimeoutExceededError(
                    "Timeout waiting for request '%s' " "to complete" % str(request)
                )

            time.sleep(delay_time)
            total_wait_time += delay_time
            delay_time = min(delay_time * 2, self._max_delay)

            request = self._easy_client.find_unique_request(id=request.id)

        if raise_on_error and request.status == "ERROR":
            raise RequestFailedError(request)

        return request

    def _get_parent_for_request(self):
        parent = getattr(request_context, "current_request", None)
        if parent is None:
            return None

        if (
            request_context.bg_host.upper() != self._bg_host.upper()
            or request_context.bg_port != self._bg_port
        ):
            self.logger.warning(
                "A parent request was found, but the destination beer-garden "
                "appears to be different than the beer-garden to which this plugin "
                "is assigned. Cross-server parent/child requests are not supported "
                "at this time. Removing the parent context so the request doesn't "
                "fail."
            )
            return None

        return Request(id=str(parent.id))

    def _construct_bg_request(self, **kwargs):
        """Create a request that can be used with EasyClient.create_request"""

        command = kwargs.pop("_command", None)
        system_name = kwargs.pop("_system_name", None)
        system_version = kwargs.pop("_system_version", None)
        system_display = kwargs.pop("_system_display", None)
        instance_name = kwargs.pop("_instance_name", None)
        comment = kwargs.pop("_comment", None)
        output_type = kwargs.pop("_output_type", None)
        metadata = kwargs.pop("_metadata", {})

        parent = self._get_parent_for_request()

        if system_display:
            metadata["system_display_name"] = system_display

        if command is None:
            raise ValidationError("Unable to send a request with no command")
        if system_name is None:
            raise ValidationError("Unable to send a request with no system name")
        if system_version is None:
            raise ValidationError("Unable to send a request with no system version")
        if instance_name is None:
            raise ValidationError("Unable to send a request with no instance name")

        return Request(
            command=command,
            system=system_name,
            system_version=system_version,
            instance_name=instance_name,
            comment=comment,
            output_type=output_type,
            parent=parent,
            metadata=metadata,
            parameters=kwargs,
        )

    @staticmethod
    def _determine_latest(systems):
        return (
            sorted(systems, key=lambda x: parse(x.version), reverse=True)[0]
            if systems
            else None
        )


class BrewmasterSystemClient(SystemClient):
    def __init__(self, *args, **kwargs):
        warnings.warn(
            "Call made to 'BrewmasterSystemClient'. This name will be removed in version "
            "3.0, please use 'SystemClient' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        super(BrewmasterSystemClient, self).__init__(*args, **kwargs)



# -*- coding: utf-8 -*-

import calendar
import datetime

import simplejson
import marshmallow
from marshmallow import Schema, post_dump, post_load, pre_load, fields
from marshmallow.utils import UTC
from marshmallow_polyfield import PolyField

__all__ = [
    "SystemSchema",
    "InstanceSchema",
    "CommandSchema",
    "ParameterSchema",
    "RequestSchema",
    "PatchSchema",
    "LoggingConfigSchema",
    "EventSchema",
    "QueueSchema",
    "PrincipalSchema",
    "RoleSchema",
    "RefreshTokenSchema",
    "JobSchema",
    "DateTriggerSchema",
    "IntervalTriggerSchema",
    "CronTriggerSchema",
]


class DateTime(fields.DateTime):
    """Class that adds methods for (de)serializing DateTime fields as an epoch"""

    def __init__(self, format="epoch", **kwargs):
        self.DATEFORMAT_SERIALIZATION_FUNCS["epoch"] = self.to_epoch
        self.DATEFORMAT_DESERIALIZATION_FUNCS["epoch"] = self.from_epoch
        super(DateTime, self).__init__(format=format, **kwargs)

    @staticmethod
    def to_epoch(dt, localtime=False):
        if localtime and dt.tzinfo is not None:
            localized = dt
        else:
            if dt.tzinfo is None:
                localized = UTC.localize(dt)
            else:
                localized = dt.astimezone(UTC)
        return (calendar.timegm(localized.timetuple()) * 1000) + int(
            localized.microsecond / 1000
        )

    @staticmethod
    def from_epoch(epoch):
        # utcfromtimestamp will correctly parse milliseconds in Python 3,
        # but in Python 2 we need to help it
        seconds, millis = divmod(epoch, 1000)
        return datetime.datetime.utcfromtimestamp(seconds).replace(
            microsecond=millis * 1000
        )


class BaseSchema(Schema):
    class Meta:
        version_nums = marshmallow.__version__.split(".")
        if int(version_nums[0]) <= 2 and int(version_nums[1]) < 17:
            json_module = simplejson
        else:
            render_module = simplejson

    def __init__(self, strict=True, **kwargs):
        super(BaseSchema, self).__init__(strict=strict, **kwargs)

    @post_load
    def make_object(self, data):
        try:
            model_class = self.context["models"][self.__class__.__name__]
        except KeyError:
            return data

        return model_class(**data)

    @classmethod
    def get_attribute_names(cls):
        return [
            key
            for key, value in cls._declared_fields.items()
            if isinstance(value, fields.FieldABC)
        ]


class ChoicesSchema(BaseSchema):

    type = fields.Str()
    display = fields.Str()
    value = fields.Raw(many=True)
    strict = fields.Bool(default=False)
    details = fields.Dict()


class ParameterSchema(BaseSchema):

    key = fields.Str()
    type = fields.Str(allow_none=True)
    multi = fields.Bool(allow_none=True)
    display_name = fields.Str(allow_none=True)
    optional = fields.Bool(allow_none=True)
    default = fields.Raw(allow_none=True)
    description = fields.Str(allow_none=True)
    choices = fields.Nested("ChoicesSchema", allow_none=True, many=False)
    parameters = fields.Nested("self", many=True, allow_none=True)
    nullable = fields.Bool(allow_none=True)
    maximum = fields.Int(allow_none=True)
    minimum = fields.Int(allow_none=True)
    regex = fields.Str(allow_none=True)
    form_input_type = fields.Str(allow_none=True)


class CommandSchema(BaseSchema):

    id = fields.Str(allow_none=True)
    name = fields.Str()
    description = fields.Str(allow_none=True)
    parameters = fields.Nested("ParameterSchema", many=True)
    command_type = fields.Str(allow_none=True)
    output_type = fields.Str(allow_none=True)
    schema = fields.Dict(allow_none=True)
    form = fields.Dict(allow_none=True)
    template = fields.Str(allow_none=True)
    icon_name = fields.Str(allow_none=True)
    system = fields.Nested("SystemSchema", only=("id",), allow_none=True)


class InstanceSchema(BaseSchema):

    id = fields.Str(allow_none=True)
    name = fields.Str()
    description = fields.Str(allow_none=True)
    status = fields.Str(allow_none=True)
    status_info = fields.Nested("StatusInfoSchema", allow_none=True)
    queue_type = fields.Str(allow_none=True)
    queue_info = fields.Dict(allow_none=True)
    icon_name = fields.Str(allow_none=True)
    metadata = fields.Dict(allow_none=True)


class SystemSchema(BaseSchema):

    id = fields.Str(allow_none=True)
    name = fields.Str()
    description = fields.Str(allow_none=True)
    version = fields.Str()
    max_instances = fields.Integer(allow_none=True)
    icon_name = fields.Str(allow_none=True)
    instances = fields.Nested("InstanceSchema", many=True, allow_none=True)
    commands = fields.Nested("CommandSchema", many=True)
    display_name = fields.Str(allow_none=True)
    metadata = fields.Dict(allow_none=True)


class RequestTemplateSchema(BaseSchema):
    """Used as a base class for request and a request template for jobs."""

    system = fields.Str(allow_none=True)
    system_version = fields.Str(allow_none=True)
    instance_name = fields.Str(allow_none=True)
    command = fields.Str(allow_none=True)
    parameters = fields.Dict(allow_none=True)
    comment = fields.Str(allow_none=True)
    metadata = fields.Dict(allow_none=True)


class RequestSchema(RequestTemplateSchema):

    id = fields.Str(allow_none=True)
    parent = fields.Nested("self", exclude=("children",), allow_none=True)
    children = fields.Nested(
        "self", exclude=("parent", "children"), many=True, default=None, allow_none=True
    )
    output = fields.Str(allow_none=True)
    output_type = fields.Str(allow_none=True)
    status = fields.Str(allow_none=True)
    command_type = fields.Str(allow_none=True)
    error_class = fields.Str(allow_none=True)
    created_at = DateTime(allow_none=True, format="epoch", example="1500065932000")
    updated_at = DateTime(allow_none=True, format="epoch", example="1500065932000")
    has_parent = fields.Bool(allow_none=True)
    requester = fields.String(allow_none=True)


class StatusInfoSchema(BaseSchema):

    heartbeat = DateTime(allow_none=True, format="epoch", example="1500065932000")


class PatchSchema(BaseSchema):

    operation = fields.Str()
    path = fields.Str(allow_none=True)
    value = fields.Raw(allow_none=True)

    @pre_load(pass_many=True)
    def unwrap_envelope(self, data, many):
        if isinstance(data, list):
            return data
        elif "operations" in data:
            return data["operations"]
        else:
            return [data]

    @post_dump(pass_many=True)
    def wrap_envelope(self, data, many):
        return {u"operations": data if many else [data]}


class LoggingConfigSchema(BaseSchema):

    level = fields.Str(allow_none=True)
    formatters = fields.Dict(allow_none=True)
    handlers = fields.Dict(allow_none=True)


class EventSchema(BaseSchema):

    name = fields.Str(allow_none=True)
    payload = fields.Dict(allow_none=True)
    error = fields.Bool(allow_none=True)
    metadata = fields.Dict(allow_none=True)
    timestamp = DateTime(allow_none=True, format="epoch", example="1500065932000")


class QueueSchema(BaseSchema):

    name = fields.Str(allow_none=True)
    system = fields.Str(allow_none=True)
    version = fields.Str(allow_none=True)
    instance = fields.Str(allow_none=True)
    system_id = fields.Str(allow_none=True)
    display = fields.Str(allow_none=True)
    size = fields.Integer(allow_none=True)


class PrincipalSchema(BaseSchema):

    id = fields.Str(allow_none=True)
    username = fields.Str(allow_none=True)
    roles = fields.Nested("RoleSchema", many=True, allow_none=True)
    permissions = fields.List(fields.Str(), allow_none=True)
    preferences = fields.Dict(allow_none=True)
    metadata = fields.Dict(allow_none=True)


class RoleSchema(BaseSchema):

    id = fields.Str(allow_none=True)
    name = fields.Str(allow_none=True)
    description = fields.Str(allow_none=True)
    roles = fields.Nested("self", many=True, allow_none=True)
    permissions = fields.List(fields.Str(), allow_none=True)


class RefreshTokenSchema(BaseSchema):

    id = fields.Str(allow_none=True)
    issued = DateTime(allow_none=True, format="epoch", example="1500065932000")
    expires = DateTime(allow_none=True, format="epoch", example="1500065932000")
    payload = fields.Dict()


class DateTriggerSchema(BaseSchema):

    run_date = DateTime(allow_none=True, format="epoch", example="1500065932000")
    timezone = fields.Str(allow_none=True)


class IntervalTriggerSchema(BaseSchema):

    weeks = fields.Int(allow_none=True)
    days = fields.Int(allow_none=True)
    hours = fields.Int(allow_none=True)
    minutes = fields.Int(allow_none=True)
    seconds = fields.Int(allow_none=True)
    start_date = DateTime(allow_none=True, format="epoch", example="1500065932000")
    end_date = DateTime(allow_none=True, format="epoch", example="1500065932000")
    timezone = fields.Str(allow_none=True)
    jitter = fields.Int(allow_none=True)
    reschedule_on_finish = fields.Bool(allow_none=True)


class CronTriggerSchema(BaseSchema):

    year = fields.Str(allow_none=True)
    month = fields.Str(allow_none=True)
    day = fields.Str(allow_none=True)
    week = fields.Str(allow_none=True)
    day_of_week = fields.Str(allow_none=True)
    hour = fields.Str(allow_none=True)
    minute = fields.Str(allow_none=True)
    second = fields.Str(allow_none=True)
    start_date = DateTime(allow_none=True, format="epoch", example="1500065932000")
    end_date = DateTime(allow_none=True, format="epoch", example="1500065932000")
    timezone = fields.Str(allow_none=True)
    jitter = fields.Int(allow_none=True)


TRIGGER_TYPE_TO_SCHEMA = {
    "interval": IntervalTriggerSchema,
    "date": DateTriggerSchema,
    "cron": CronTriggerSchema,
}


def serialize_trigger_selector(_, obj):
    try:
        return TRIGGER_TYPE_TO_SCHEMA[obj.trigger_type]()
    except KeyError:
        pass

    raise TypeError("Could not detect %s trigger type schema" % obj.trigger_type)


def deserialize_trigger_selector(_, data):
    try:
        return TRIGGER_TYPE_TO_SCHEMA[data["trigger_type"]]()
    except KeyError:
        pass

    raise TypeError("Could not detect %s trigger type schema" % data["trigger_type"])


class JobSchema(BaseSchema):

    id = fields.Str(allow_none=True)
    name = fields.Str(allow_none=True)
    trigger_type = fields.Str(allow_none=True)
    trigger = PolyField(
        allow_none=True,
        serialization_schema_selector=serialize_trigger_selector,
        deserialization_schema_selector=deserialize_trigger_selector,
    )
    request_template = fields.Nested("RequestTemplateSchema")
    misfire_grace_time = fields.Int(allow_none=True)
    coalesce = fields.Bool(allow_none=True)
    next_run_time = DateTime(allow_none=True, format="epoch", example="1500065932000")
    success_count = fields.Int(allow_none=True)
    error_count = fields.Int(allow_none=True)
    status = fields.Str(allow_none=True)
    max_instances = fields.Int(allow_none=True)



# -*- coding: utf-8 -*-

SPECIFICATION = {
    "bg_host": {
        "type": "str",
        "description": "The beergarden server FQDN",
        "required": True,
        "env_name": "HOST",
        "alt_env_names": ["WEB_HOST"],
    },
    "bg_port": {
        "type": "int",
        "description": "The beergarden server port",
        "default": 2337,
        "env_name": "PORT",
        "alt_env_names": ["WEB_PORT"],
    },
    "ca_cert": {
        "type": "str",
        "description": "CA certificate to use when verifying",
        "required": False,
        "alt_env_names": ["SSL_CA_CERT"],
    },
    "ca_verify": {
        "type": "bool",
        "description": "Verify server certificate when using SSL",
        "default": True,
    },
    "client_cert": {
        "type": "str",
        "description": "Client certificate to use with beergarden",
        "required": False,
        "alt_env_names": ["SSL_CLIENT_CERT"],
    },
    "ssl_enabled": {
        "type": "bool",
        "description": "Use SSL when communicating with beergarden",
        "default": True,
    },
    "url_prefix": {
        "type": "str",
        "description": "The beergarden server path",
        "default": "/",
    },
    "api_version": {
        "type": "int",
        "description": "Beergarden API version",
        "required": False,
    },
    "username": {
        "type": "str",
        "description": "Username for authentication",
        "required": False,
    },
    "password": {
        "type": "str",
        "description": "Password for authentication",
        "required": False,
    },
    "access_token": {
        "type": "str",
        "description": "Access token for authentication",
        "required": False,
    },
    "refresh_token": {
        "type": "str",
        "description": "Refresh token for authentication",
        "required": False,
    },
    "client_timeout": {
        "type": "float",
        "description": "Max time RestClient will wait for server response",
        "long_description": "This setting controls how long the HTTP(s) client will wait "
        "when opening a connection to Beergarden before aborting."
        "This prevents some strange Beergarden server state from causing "
        "plugins to hang indefinitely."
        "Set to -1 to disable (this is a bad idea in production code, see "
        "the Requests documentation).",
        "default": -1,
    },
}
